import torch as th
import math
import numpy as np
from video_loader import VideoLoader
from torch.utils.data import DataLoader
import argparse
from model import get_model
from preprocessing import Preprocessing
from random_sequence_shuffler import RandomSequenceSampler
import torch.nn.functional as F
import os
parser = argparse.ArgumentParser(description='Easy video feature extractor')

parser.add_argument(
    '--csv',
    type=str,
    help='input csv with video input path')
parser.add_argument('--batch_size', type=int, default=64,
                            help='batch size')
parser.add_argument('--type', type=str, default='2d',
                            help='CNN type')
parser.add_argument('--half_precision', type=int, default=1,
                            help='output half precision float')
parser.add_argument('--num_decoding_thread', type=int, default=4,
                            help='Num parallel thread for video decoding')
parser.add_argument('--l2_normalize', type=int, default=1,
                            help='l2 normalize feature')
parser.add_argument('--resnext101_model_path', type=str, default='model/resnext101.pth',
                            help='Resnext model path')
args = parser.parse_args()

if __name__ == '__main__':
    dataset = VideoLoader(
        args.csv,
        framerate=1 if args.type == '2d' else 24,
        size=224 if args.type == '2d' else 112,
        centercrop=(args.type == '3d'),
    )
    n_dataset = len(dataset)
    sampler = RandomSequenceSampler(n_dataset, 10)
    loader = DataLoader(
        dataset,
        batch_size=1,
        shuffle=False,
        num_workers=args.num_decoding_thread,
        sampler=sampler if n_dataset > 10 else None,
    )
    preprocess = Preprocessing(args.type)
    model = get_model(args)
    print()
    with th.no_grad():
        for k, data in enumerate(loader):
            input_file = data['input'][0]
            if not os.path.exists(input_file):
                print('file is not exists')
                raise("error" )
            print(input_file)
            output_file = data['output'][0]
            print(data["video"].shape)
            print(len(data['video'].shape) )
            if len(data['video'].shape) > 3:
                print('Computing features of video {}/{}: {}'.format(
                    k + 1, n_dataset, input_file))
                video = data['video'].squeeze()
                print("len(video.shape)=", len(video.shape))
                if len(video.shape) == 4:
                    video = preprocess(video)
                    n_chunk = len(video)
                    print("n_chunk = ",n_chunk)
                    print("batch size = ",args.batch_size)
                    features = th.cuda.FloatTensor(n_chunk, 2048).fill_(0)
                    n_iter = int(math.ceil(n_chunk / float(args.batch_size)))
                    print("number of inter. = ",n_iter)
                    for i in range(n_iter):
                        min_ind = i * args.batch_size
                        max_ind = (i + 1) * args.batch_size
                        video_batch = video[min_ind:max_ind].cuda()
                        batch_features = model(video_batch)
                        if args.l2_normalize:
                            batch_features = F.normalize(batch_features, dim=1)
                        features[min_ind:max_ind] = batch_features
                    features = features.cpu().numpy()
                    print("features.shape = ",features.shape)
                    if args.half_precision:
                        features = features.astype('float16')
                    print("save...", output_file)
                    print(features.shape)
                    np.save(output_file, features)
                    print()
            else:
                print('Video {} already processed.'.format(input_file))
                print()




import os
import pathlib
from typing import Dict, Union

import numpy as np
import torch
from tqdm import tqdm
import traceback

from mmkfeatures.video.utils.utils import form_list_from_user_input, extract_wav_from_mp4, action_on_extraction
from mmkfeatures.video.models.vggish.vggish_src.vggish_slim import VGGish


class ExtractVGGish(torch.nn.Module):

    def __init__(self, args):
        super(ExtractVGGish, self).__init__()
        self.feature_type = args.feature_type
        self.path_list = form_list_from_user_input(args)
        self.keep_tmp_files = args.keep_tmp_files
        self.on_extraction = args.on_extraction
        self.tmp_path = os.path.join(args.tmp_path, self.feature_type)
        self.output_path = os.path.join(args.output_path, self.feature_type)
        self.progress = tqdm(total=len(self.path_list))
        if args.show_pred:
            raise NotImplementedError

    def forward(self, indices: torch.LongTensor):
        '''
        Arguments:
            indices {torch.LongTensor} -- indices to self.path_list
        '''
        device = indices.device
        model = self.load_model(device)

        for idx in indices:
            # when error occurs might fail silently when run from torch data parallel
            try:
                feats_dict = self.extract(device, model, self.path_list[idx])
                action_on_extraction(feats_dict, self.path_list[idx], self.output_path, self.on_extraction)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except Exception as e:
                # prints only the last line of an error. Use `traceback.print_exc()` for the whole traceback
                traceback.print_exc()
                print(e)
                print(f'Extraction failed at: {self.path_list[idx]} with error (↑). Continuing extraction')

            # update tqdm progress bar
            self.progress.update()

    def extract(self,
                device: torch.device,
                model: torch.nn.Module,
                video_path: Union[str, None] = None) -> Dict[str, np.ndarray]:
        '''The extraction call. Made to clean the forward call a bit.

        Args:
            device (torch.device): the device
            model (torch.nn.Module): the model
            video_path (Union[str, None], optional): Path to a video. Defaults to None.

        Keyword Arguments:
            video_path {Union[str, None]} -- if you would like to use import it and use it as
                                             "path -> model"-fashion (default: {None})

        Returns:
            Dict[str, np.ndarray]: extracted VGGish features
        '''
        file_ext = pathlib.Path(video_path).suffix

        if file_ext == '.mp4':
            # extract audio files from .mp4
            audio_wav_path, audio_aac_path = extract_wav_from_mp4(video_path, self.tmp_path)
        elif file_ext == '.wav':
            audio_wav_path = video_path
            audio_aac_path = None
        else:
            raise NotImplementedError

        with torch.no_grad():
            vggish_stack = model(audio_wav_path, device).cpu().numpy()

        # removes the folder with audio files created during the process
        if not self.keep_tmp_files:
            if video_path.endswith('.mp4'):
                os.remove(audio_wav_path)
                os.remove(audio_aac_path)

        feats_dict = {self.feature_type: vggish_stack}

        return feats_dict

    def load_model(self, device: torch.device) -> torch.nn.Module:
        '''Defines the models, loads checkpoints, sends them to the device.

        Args:
            device (torch.device)

        Returns:
            torch.nn.Module: the model
        '''
        model = VGGish()
        model = model.to(device)
        model.eval()
        return model



# import the necessary packages
import numpy as np
import csv
class Searcher:
	def __init__(self, indexPath):
		# store our index path
		self.indexPath = indexPath

	def search(self, queryFeatures, limit = 10):
		# initialize our dictionary of results
		results = {}

		# open the index file for reading
		with open(self.indexPath) as f:
			# initialize the CSV reader
			reader = csv.reader(f)
			# loop over the rows in the index
			for row in reader:
				# parse out the image ID and features, then compute the
				# chi-squared distance between the features in our index
				# and our query features
				features = [float(x) for x in row[1:]]
				d = self.chi2_distance(features, queryFeatures)
				# now that we have the distance between the two feature
				# vectors, we can udpate the results dictionary -- the
				# key is the current image ID in the index and the
				# value is the distance we just computed, representing
				# how 'similar' the image in the index is to our query
				results[row[0]] = d
			# close the reader
			f.close()
		# sort our results, so that the smaller distances (i.e. the
		# more relevant images are at the front of the list)
		results = sorted([(v, k) for (k, v) in results.items()])
		# return our (limited) results
		return results[:limit]

	def chi2_distance(self, histA, histB, eps = 1e-10):
		# compute the chi-squared distance
		d = 0.5 * np.sum([((a - b) ** 2) / (a + b + eps)
			for (a, b) in zip(histA, histB)])
		# return the chi-squared distance
		return d


import torch
torch.cuda.set_per_process_memory_fraction(1.0, 0)
import torch.nn as nn
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
from mmkfeatures.models.image_text_fusion.conv_autoencoder.models import ConvAutoencoder, ImgDataset
import argparse
import time
import os


def ensure_folder(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)


def time_msec():
    return int(round(time.time() * 1000))


def train(input_folder, output_folder, lr=0.001, epochs=40, img_format='.jpeg'):
    ensure_folder(output_folder)
    model = ConvAutoencoder()
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    transform = transforms.ToTensor()
    train_folder = input_folder + '/train'
    print(f"Create train dataset from - {train_folder}")
    train_data = ImgDataset(train_folder, transform, device, img_format)
    print(f"Train dataset size - {len(train_data)}")
    # Create training
    num_workers = 0
    # how many samples per batch to load
    batch_size = 2
    # prepare data loaders
    print(f"Create dataloader batches={batch_size} and workers={num_workers}")
    train_loader = DataLoader(train_data, batch_size=batch_size, num_workers=num_workers, shuffle=False)
    print("trainloader's len:",len(train_loader))
    # specify loss function
    criterion = nn.BCELoss()
    # specify loss function
    optimizer = torch.optim.Adam(model.parameters(), lr)
    # number of epochs to train the model
    n_epochs = epochs

    for epoch in range(1, n_epochs + 1):
        # monitor training loss
        train_loss = 0.0

        ###################
        # Train the model #
        ###################
        model.train()
        train_loss = 0
        for images in train_loader:
            # _ stands in for labels, here
            # no need to flatten enc_64x64_images
            # clear the gradients of all optimized variables
            optimizer.zero_grad()
            # forward pass: compute predicted outputs by passing inputs to the model
            outputs, encoder_output = model(images, encoder_mode=False)
            # calculate the loss
            loss = criterion(outputs, images)
            # backward pass: compute gradient of the loss with respect to model parameters
            loss.backward()
            # perform a single optimization step (parameter update)
            optimizer.step()
            # update running training loss
            train_loss += loss.item() * images.size(0)

        # print avg training statistics
        train_loss = train_loss / len(train_loader)
        print('Epoch: {} \tTraining Loss: {:.6f}'.format(epoch, train_loss))

    ###################
    # Test the model #
    ###################
    model.eval()
    test_folder = input_folder + '/test'
    test_data = ImgDataset(test_folder, transform, device, img_format)
    test_loader = DataLoader(test_data, batch_size=batch_size, num_workers=num_workers)
    eval_loss = 0
    for images in test_loader:
        # forward pass: compute predicted outputs by passing inputs to the model
        outputs, x_encoders = model(images, encoder_mode=False)
        # calculate the loss
        loss = criterion(outputs, images)
        # update running training loss
        eval_loss += loss.item() * images.size(0)

    # print avg evaluation statistics
    eval_loss = eval_loss / len(test_loader)
    print('\tEvaluation Loss: {:.6f}'.format(eval_loss))

    # Save model
    # model_root = output_folder + '/conv_autoencoder_' + str(time_msec()) + '.pt'
    model_root = output_folder + '/conv_autoencoder.pt'
    print(f"Saving trained model to: {model_root}")
    stat_dict=model.state_dict()
    torch.save(stat_dict, model_root)


def main():
    '''
train.py
../../dataset
../../output
0.001
100
    '''
    parser = argparse.ArgumentParser(description='Train conv autoencoder on dataset.'
                                                 'Must have train and test dataset folders'
                                                 ' located in input')
    parser.add_argument('input', metavar='input', type=str,
                        help='Source folder full path.')
    parser.add_argument('output', metavar='output', type=str,
                        help='Destination folder full path.')
    parser.add_argument('lr', metavar='lt', type=str,
                        help='Learning rate')
    parser.add_argument('epochs', metavar='epochs', type=str,
                        help='Number of training epochs.')

    args = parser.parse_args()
    input_folder = args.input
    output_folder = args.output
    lr = args.lr
    epochs = args.epochs

    train(input_folder, output_folder, float(lr), int(epochs))


if __name__ == '__main__':
    print("<<<<<<<<<<<<<<< Start training >>>>>>>>>>>>>>>>>")
    main()



import csv
import os
import sys
from abc import ABC

import math
import numpy as np
from keras.preprocessing import image as image_preprocessor


class Extractor(ABC):
    
    def __init__(self, base_route: str, size=224, batch_size=128):
        
        self.base_route: str = base_route
        self.output_file: str = ''
        if os.path.isdir(base_route):
            self.num_classes: int = len(os.listdir(base_route))
        else:
            self.num_classes = 1
        self.file_writer = None
        
        self.bath_size = batch_size
        self.__width = self.__height = size
        self.image_shape = (self.width, self.height, 3)
        self.directory_iterator = None
        
        self.counter: int = 0
        self.verbose: bool = True
        self.header: int = 1
        self.verbosity_frequency: int = 10
        if os.path.isdir(self.base_route):
            self._set_directory_iterator(self.base_route)

        super().__init__()
    
    def extract_and_save(self, output_file: str, verbose: bool = True, verbosity_frequency: int = 10, header: int = 1):
        self.output_file = output_file
        self.header: int = header
        self.verbose: bool = verbose
        self.verbosity_frequency: int = verbosity_frequency
        
        self._reset_counter()
        with open(self.output_file, 'w') as f:
            self.file_writer = csv.writer(f)
            self._print_header()
            
            for filename, category in zip(self.directory_iterator.filenames, self.directory_iterator.classes):
                image_route = os.path.join(self.base_route, filename)
                image_route=image_route.replace("\\","/")
                features = self.extract(image_route=image_route)
                self._print_row(features=features, label=category)
                self._update_counter()
    
    def extract(self, image_route: str) -> np.ndarray:
        pass
    
    def _set_directory_iterator(self, route: str) -> None:
        image_generator = image_preprocessor.ImageDataGenerator(rescale=1.0 / 255)
        
        self.directory_iterator = image_generator.flow_from_directory(
            directory=route,
            target_size=(self.width, self.height),
            batch_size=self.bath_size,
            class_mode="categorical"
        )
    
    def _print_row(self, features: np.ndarray, label) -> None:
        self.file_writer.writerow(np.append(features, [label]))
    
    def _print_header(self):
        if self.header:
            column_names = list(range(self._find_features_size())) + [-1]
            self.file_writer.writerow(column_names)
    
    def _find_features_size(self) -> int:
        pass
    
    def _update_counter(self) -> None:
        self.counter += 1
        
        if self.verbose and ((self.counter % self.verbosity_frequency) == 0):
            progress_percent: int = math.floor((self.counter / self.num_classes) * 100)
            # sys.stdout.write("{}%\n".format(progress_percent))
    
    def _reset_counter(self) -> None:
        self.counter = 0
    
    @property
    def width(self):
        return self.__width
    
    @width.setter
    def width(self, width):
        self.__width = width
        self.image_shape = (width, self.height, 3)
    
    @property
    def height(self):
        return self.__height
    
    @height.setter
    def height(self, height):
        self.__height = height
        self.image_shape = (self.width, height, 3)



# we import unused `numpy` before `torch` because if run from `subprocess.run()`
# it fails with
# `Error: mkl-service + Intel(R) MKL: MKL_THREADING_LAYER=INTEL is incompatible with libgomp.so.1 library.`
# (see https://github.com/pytorch/pytorch/issues/37377)
import numpy
import torch
import os
from omegaconf import OmegaConf

from mmkfeatures.video.utils.utils import build_cfg_path, form_list_from_user_input, sanity_check

def parallel_feature_extraction(args):
    '''Distributes the feature extraction in embarasingly-parallel fashion. Specifically,
    it divides the dataset (list of video paths) among all specified devices evenly and extract features.'''

    if args.feature_type == 'i3d':
        from mmkfeatures.video.models.i3d.extract_i3d import ExtractI3D  # defined here to avoid import errors
        extractor = ExtractI3D(args)
        pass
    elif args.feature_type == 'r21d':
        from mmkfeatures.video.models.r21d.extract_r21d import ExtractR21D  # defined here to avoid import errors
        extractor = ExtractR21D(args)
    elif args.feature_type == 'vggish':
        from mmkfeatures.video.models.vggish.extract_vggish import ExtractVGGish  # defined here to avoid import errors
        extractor = ExtractVGGish(args)
        pass
    elif args.feature_type in ['resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']:
        from mmkfeatures.video.models.resnet.extract_resnet import ExtractResNet
        extractor = ExtractResNet(args)
    elif args.feature_type == 'raft':
        from mmkfeatures.video.models.raft.extract_raft import ExtractRAFT
        extractor = ExtractRAFT(args)
        pass
    elif args.feature_type == 'pwc':
        from mmkfeatures.video.models.pwc.extract_pwc import ExtractPWC
        extractor = ExtractPWC(args)
        pass
    else:
        raise NotADirectoryError

    # the indices correspond to the positions of the target videos in
    # the video_paths list. They are required here because
    # scatter module inputs only tensors but there is no such torch tensor
    # that would be suitable for strings (video_paths). Also, the
    # input have the method '.device' which allows us to access the
    # current device in the extractor.
    video_paths = form_list_from_user_input(args)
    indices = torch.arange(len(video_paths))
    replicas = torch.nn.parallel.replicate(extractor, args.device_ids[:len(indices)])
    inputs = torch.nn.parallel.scatter(indices, args.device_ids[:len(indices)])
    torch.nn.parallel.parallel_apply(replicas[:len(inputs)], inputs)
    # closing the tqdm progress bar to avoid some unexpected errors due to multi-threading
    extractor.progress.close()

def extract_video_features_by_model(video_path,feature_type):

    cfg_cli = OmegaConf.from_cli()
    cfg_cli.feature_type = feature_type
    cfg_cli.video_paths = [video_path]
    print(cfg_cli)
    config_path = os.path.dirname(os.path.realpath(__file__))+"/configs"
    cfg_yml = OmegaConf.load(build_cfg_path(feature_type=cfg_cli.feature_type,root_configs=config_path))
    # the latter arguments are prioritized
    cfg = OmegaConf.merge(cfg_yml, cfg_cli)
    # OmegaConf.set_readonly(cfg, True)
    print(OmegaConf.to_yaml(cfg))
    # some printing
    if cfg.on_extraction in ['save_numpy', 'save_pickle']:
        print(f'Saving features to {cfg.output_path}')
    if cfg.keep_tmp_files:
        print(f'Keeping temp files in {cfg.tmp_path}')

    sanity_check(cfg)
    parallel_feature_extraction(cfg)

if __name__ == "__main__":
    cfg_cli = OmegaConf.from_cli()
    print(cfg_cli)
    cfg_yml = OmegaConf.load(build_cfg_path(cfg_cli.feature_type))
    # the latter arguments are prioritized
    cfg = OmegaConf.merge(cfg_yml, cfg_cli)
    # OmegaConf.set_readonly(cfg, True)
    print(OmegaConf.to_yaml(cfg))
    # some printing
    if cfg.on_extraction in ['save_numpy', 'save_pickle']:
        print(f'Saving features to {cfg.output_path}')
    if cfg.keep_tmp_files:
        print(f'Keeping temp files in {cfg.tmp_path}')

    sanity_check(cfg)
    parallel_feature_extraction(cfg)



import pandas as pd
from deepface import DeepFace

class VideoEmotionToolkit:

    def __init__(self):
        self.elist=[]
        self.count=0
        pass

    def empty_result(self):
        self.elist=[]
        self.count=0

    def call_back(self,emotions,age,gender,timepoint):
        dict_model={}
        for index, row in emotions.iterrows():
            dict_model[str(row["emotion"]).lower()]=row["score"]
        interval=(self.count*self.time_threshold,(self.count+1)*self.time_threshold)
        self.elist.append([timepoint,dict_model])
        self.count+=1

    def get_result(self):
        return self.elist

    def analyze(self,source,time_thre=5):
        self.time_threshold=time_thre
        DeepFace.stream(
            # db_path="../doctor_images",
            # source=0,
            source=source,
            enable_face_recognition=False, call_back_func=self.call_back, time_threshold=time_thre)


if __name__=="__main__":
    video_toolkit = VideoEmotionToolkit()
    video_toolkit.empty_result()

    path = "deep_face_analysis/video/test.mp4"
    video_toolkit.analyze(path, time_thre=5)
    result = video_toolkit.get_result()

    print()
    for item in result:
        print(item[0], item[1])






import torch
import time
from torch import nn
import torch.nn.functional as F


class LSTHM(nn.Module):

	def __init__(self,cell_size,in_size,hybrid_in_size):
		super(LSTHM, self).__init__()
		self.cell_size=cell_size
		self.in_size=in_size
		self.W=nn.Linear(in_size,4*self.cell_size)
		self.U=nn.Linear(cell_size,4*self.cell_size)
		self.V=nn.Linear(hybrid_in_size,4*self.cell_size)

	def step(self,x,ctm1,htm1,ztm1):
		input_affine=self.W(x)
		output_affine=self.U(htm1)
		hybrid_affine=self.V(ztm1)
		
		sums=input_affine+output_affine+hybrid_affine

		#biases are already part of W and U and V
		f_t=F.sigmoid(sums[:,:self.cell_size])
		i_t=F.sigmoid(sums[:,self.cell_size:2*self.cell_size])
		o_t=F.sigmoid(sums[:,2*self.cell_size:3*self.cell_size])
		ch_t=F.tanh(sums[:,3*self.cell_size:])
		c_t=f_t*ctm1+i_t*ch_t
		h_t=F.tanh(c_t)*o_t
		return c_t,h_t

	



import torch
import torch.nn as nn
from mmkfeatures.models.image_text_fusion.char_embedding.fixed_gru import FixedGru
from mmkfeatures.models.image_text_fusion.char_embedding.fixed_rnn import FixedRnn


class CharCnnRnn(nn.Module):
    # Char-cnn-rnn text embedding
    def __init__(self, rnn_type='fixed_rnn', model_type='cvpr'):
        super().__init__()
        if model_type == 'cvpr':
            rnn_dim = 256
            use_maxpool3 = True
            rnn = FixedRnn
            rnn_num_steps = 8
        else:
            # icml model type
            rnn_dim = 512
            if rnn_type == 'fixed_rnn':
                use_maxpool3 = True
                rnn = FixedRnn
                rnn_num_steps = 8
            else:
                # Use fixed_gru
                use_maxpool3 = False
                rnn = FixedGru
                rnn_num_steps = 18

        self.rnn_type = rnn_type
        self.model_type = model_type
        self.use_maxpool3 = use_maxpool3

        # network setup
        # (B, 70, 201)
        self.conv1 = nn.Conv1d(70, 384, kernel_size=4)
        self.threshold1 = nn.Threshold(1e-6, 0)
        self.maxpool1 = nn.MaxPool1d(kernel_size=3, stride=3)
        # (B, 384, 66)
        self.conv2 = nn.Conv1d(384, 512, kernel_size=4)
        self.threshold2 = nn.Threshold(1e-6, 0)
        self.maxpool2 = nn.MaxPool1d(kernel_size=3, stride=3)
        # (B, 512, 21)
        self.conv3 = nn.Conv1d(512, rnn_dim, kernel_size=4)
        self.threshold3 = nn.Threshold(1e-6, 0)
        if use_maxpool3:
            self.maxpool3 = nn.MaxPool1d(kernel_size=3, stride=2)
        # (B, rnn_dim, rnn_num_steps)
        self.rnn = rnn(num_steps=rnn_num_steps, emb_dim=rnn_dim)
        # (B, rnn_dim)
        self.emb_proj = nn.Linear(rnn_dim, 1024)
        # (B, 1024)

    def forward(self, txt):
        # temporal convolutions
        out = self.conv1(txt)
        out = self.threshold1(out)
        out = self.maxpool1(out)

        out = self.conv2(out)
        out = self.threshold2(out)
        out = self.maxpool2(out)

        out = self.conv3(out)
        out = self.threshold3(out)
        if self.use_maxpool3:
            out = self.maxpool3(out)

        # recurrent computation
        out = out.permute(0, 2, 1)
        out = self.rnn(out)

        # linear projection
        out = self.emb_proj(out)

        return out


def str_to_labelvec(string, max_str_len=201):
    string = string.lower()
    alphabet = "abcdefghijklmnopqrstuvwxyz0123456789-,;.!?:'\"/\\|_@#$%^&*~`+-=<>()[]{} "
    # {'char': num, ...}
    alpha_to_num = {k: v + 1 for k, v in zip(alphabet, range(len(alphabet)))}
    labels = torch.zeros(max_str_len, requires_grad=False).long()
    max_i = min(max_str_len, len(string))
    for i in range(max_i):
        # Append ' ' number if char not found
        labels[i] = alpha_to_num.get(string[i], alpha_to_num[' '])
    return labels


def labelvec_to_onehot(labels):
    labels = torch.LongTensor(labels).unsqueeze(1)
    one_hot = torch.zeros(labels.size(0), 71, requires_grad=False).scatter_(1, labels, 1.)
    # Ignore zeros in one-hot mask (position 0 = empty one-hot)model.state_dict()
    one_hot = one_hot[:, 1:]
    one_hot = one_hot.permute(1, 0)
    return one_hot


def prepare_text(string, max_str_len=201):
    # Converts a text description from string format to one-hot tensor format.
    labels = str_to_labelvec(string, max_str_len)
    one_hot = labelvec_to_onehot(labels)
    return one_hot



import os
from typing import Dict, Tuple, Union

import cv2
import numpy as np
import torch
import torchvision.models as models
import torchvision.transforms as transforms
from tqdm import tqdm
from mmkfeatures.video.utils.utils import (action_on_extraction, form_list_from_user_input,
                         reencode_video_with_diff_fps,
                         show_predictions_on_dataset)

# import traceback


RESIZE_SIZE = 256
CENTER_CROP_SIZE = 224
TRAIN_MEAN = [0.485, 0.456, 0.406]
TRAIN_STD = [0.229, 0.224, 0.225]


class ExtractResNet(torch.nn.Module):

    def __init__(self, args):
        super(ExtractResNet, self).__init__()
        self.feature_type = args.feature_type
        self.path_list = form_list_from_user_input(args)
        self.batch_size = args.batch_size
        self.central_crop_size = CENTER_CROP_SIZE
        self.extraction_fps = args.extraction_fps
        self.transforms = transforms.Compose([
            transforms.ToPILImage(),
            transforms.Resize(RESIZE_SIZE),
            transforms.CenterCrop(CENTER_CROP_SIZE),
            transforms.ToTensor(),
            transforms.Normalize(mean=TRAIN_MEAN, std=TRAIN_STD)
        ])
        self.show_pred = args.show_pred
        # not used, create an issue if you would like to save the frames
        self.keep_tmp_files = args.keep_tmp_files
        self.on_extraction = args.on_extraction
        self.tmp_path = os.path.join(args.tmp_path, self.feature_type)
        self.output_path = os.path.join(args.output_path, self.feature_type)
        self.progress = tqdm(total=len(self.path_list))

    def forward(self, indices: torch.LongTensor):
        '''
        Arguments:
            indices {torch.LongTensor} -- indices to self.path_list
        '''
        device = indices.device
        model, class_head = self.load_model(device)

        for idx in indices:
            # when error occurs might fail silently when run from torch data parallel
            try:
                feats_dict = self.extract(device, model, class_head, self.path_list[idx])
                action_on_extraction(feats_dict, self.path_list[idx], self.output_path, self.on_extraction)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except Exception as e:
                # prints only the last line of an error. Use `traceback.print_exc()` for the whole traceback
                print(e)
                print(f'Extraction failed at: {self.path_list[idx]} with error (↑). Continuing extraction')

            # update tqdm progress bar
            self.progress.update()

    def extract(self, device: torch.device, model: torch.nn.Module, classifier: torch.nn.Module,
                video_path: Union[str, None] = None) -> Dict[str, np.ndarray]:
        '''The extraction call. Made to clean the forward call a bit.

        Arguments:
            device {torch.device}
            model {torch.nn.Module}
            classifier {torch.nn.Module} -- pre-trained classification layer, will be used if
                                            show_pred is True

        Keyword Arguments:
            video_path {Union[str, None]} -- if you would like to use import it and use it as
                                             "path -> model"-fashion (default: {None})

        Returns:
            Dict[str, np.ndarray]: 'features_nme', 'fps', 'timestamps_ms'
        '''
        def _run_on_a_batch(vid_feats, batch, model, classifier, device):
            batch = torch.cat(batch).to(device)

            with torch.no_grad():
                batch_feats = model(batch)
                vid_feats.extend(batch_feats.tolist())
                # show predicitons on imagenet dataset (might be useful for debugging)
                if self.show_pred:
                    logits = classifier(batch_feats)
                    show_predictions_on_dataset(logits, 'imagenet')

        # take the video, change fps and save to the tmp folder
        if self.extraction_fps is not None:
            video_path = reencode_video_with_diff_fps(video_path, self.tmp_path, self.extraction_fps)

        # read a video
        cap = cv2.VideoCapture(video_path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        timestamps_ms = []
        batch = []
        vid_feats = []

        # sometimes when the target fps is 1 or 2, the first frame of the reencoded video is missing
        # and cap.read returns None but the rest of the frames are ok. timestep is 0.0 for the 2nd frame in
        # this case
        first_frame = True
        while cap.isOpened():
            frame_exists, rgb = cap.read()

            if first_frame:
                first_frame = False
                if frame_exists is False:
                    continue

            if frame_exists:
                timestamps_ms.append(cap.get(cv2.CAP_PROP_POS_MSEC))
                # prepare data (first -- transform, then -- unsqueeze)
                rgb = cv2.cvtColor(rgb, cv2.COLOR_BGR2RGB)
                rgb = self.transforms(rgb)
                rgb = rgb.unsqueeze(0)
                batch.append(rgb)
                # when batch is formed to inference
                if len(batch) == self.batch_size:
                    _run_on_a_batch(vid_feats, batch, model, classifier, device)
                    # clean up the batch list
                    batch = []
            else:
                # if the last batch was smaller than the batch size
                if len(batch) != 0:
                    _run_on_a_batch(vid_feats, batch, model, classifier, device)
                cap.release()
                break

        # removes the video with different fps if it was created to preserve disk space
        if (self.extraction_fps is not None) and (not self.keep_tmp_files):
            os.remove(video_path)

        features_with_meta = {
            self.feature_type: np.array(vid_feats),
            'fps': np.array(fps),
            'timestamps_ms': np.array(timestamps_ms)
        }

        return features_with_meta

    def load_model(self, device: torch.device) -> Tuple[torch.nn.Module]:
        '''Defines the models, loads checkpoints, sends them to the device.

        Args:
            device (torch.device): The device

        Raises:
            NotImplementedError: if flow type is not implemented.

        Returns:
            Tuple[torch.nn.Module]: the model with identity head, the original classifier
        '''
        if self.feature_type == 'resnet18':
            model = models.resnet18
        elif self.feature_type == 'resnet34':
            model = models.resnet34
        elif self.feature_type == 'resnet50':
            model = models.resnet50
        elif self.feature_type == 'resnet101':
            model = models.resnet101
        elif self.feature_type == 'resnet152':
            model = models.resnet152
        else:
            raise NotImplementedError

        model = model(pretrained=True)
        model = model.to(device)
        model.eval()
        # save the pre-trained classifier for show_preds and replace it in the net with identity
        class_head = model.fc
        model.fc = torch.nn.Identity()
        return model, class_head



import math
import numpy as np

# method 1: absSum
def calVolume(waveData, frameSize, overLap):
    wlen = len(waveData)
    step = frameSize - overLap
    frameNum = int(math.ceil(wlen*1.0/step))
    volume = np.zeros((frameNum,1))
    for i in range(frameNum):
        curFrame = waveData[np.arange(i*step,min(i*step+frameSize,wlen))]
        #curFrame = curFrame - np.median(curFrame) # False
        curFrame = curFrame - np.mean(curFrame) # zero-justified
        volume[i] = np.sum(np.abs(curFrame))
    return volume

# method 2: log10 of square sum
def calVolumeDB(waveData, frameSize, overLap):
    wlen = len(waveData)
    step = frameSize - overLap
    frameNum = int(math.ceil(wlen*1.0/step))
    volume = np.zeros((frameNum,1))
    for i in range(frameNum):
        curFrame = waveData[np.arange(i*step,min(i*step+frameSize,wlen))]
        curFrame = curFrame - np.mean(curFrame) # zero-justified
        volume[i] = 10*np.log10(np.sum(curFrame*curFrame))
    return volume



import random
import pickle
import matplotlib
matplotlib.use("Agg")
from mmkfeatures.image.conv_autoencoder import ConvAutoencoder
from tensorflow.keras.optimizers import Adam
import matplotlib.pyplot as plt
from tensorflow.keras.models import Model
from mmkfeatures.fusion.mm_features_lib import MMFeaturesLib
import numpy as np
import cv2
from tqdm import tqdm
from tensorflow.keras.models import load_model
from imutils import build_montages

class ImageAutoEncoderBuilder():

    def __init__(self,epochs=20,init_lr=1e-3,bs=32,img_size=(64,64),channel=3,split=0.8):
        self.EPOCHS = epochs
        self.INIT_LR = init_lr
        self.BS = bs
        self.img_size = img_size
        self.channel = channel
        self.split = split
        self.H=None
        self.autoencoder=None

    def create_datasets(self, data, is_shuffle=False):
        list_cids = list(data.keys())
        if is_shuffle:
            random.shuffle(list_cids)
        Xs=[]
        Ids=[]
        for cid in tqdm(list_cids):
            item = data[cid]
            imgs = item["objects"]
            for img_id in imgs:
                # print(image)
                img = imgs[img_id][()]
                resized_img = cv2.resize(img, dsize=self.img_size)
                Xs.append(resized_img)
                Ids.append(cid)
        Xs = np.array(Xs)
        Xs = Xs.astype("float32") / 255.0
        return Ids,Xs

    def create_train_test_ids(self,data,total_sample_size=-1, is_shuffle=False):
        print("[INFO] loading dataset...")
        trainX = []
        testX = []

        list_cids = list(data.keys())

        if is_shuffle:
            random.shuffle(list_cids)

        if total_sample_size != -1:
            list_cids = list_cids[:total_sample_size]

        if is_shuffle:
            random.shuffle(list_cids)

        total_num = len(list_cids)
        train_num = int(total_num * self.split)

        test_num = total_num - train_num

        list_cids_train = list_cids[:train_num]
        list_cids_test = list_cids[train_num:]

        for cid in tqdm(list_cids_train):
            item = data[cid]
            imgs = item["objects"]
            for img_id in imgs:
                # print(image)
                img = imgs[img_id][()]
                resized_img = cv2.resize(img, dsize=self.img_size)
                trainX.append(resized_img)

        for cid in tqdm(list_cids_test):
            item = data[cid]
            imgs = item["objects"]
            for img_id in imgs:
                # print(image)
                img = imgs[img_id][()]
                resized_img = cv2.resize(img, dsize=self.img_size)
                testX.append(resized_img)

        trainX = np.array(trainX)
        testX = np.array(testX)
        trainX = trainX.astype("float32") / 255.0
        testX = testX.astype("float32") / 255.0
        return trainX,testX,list_cids_train,list_cids_test

    def train(self,trainX,testX, output_model_path=""):

        # construct our convolutional autoencoder
        print("[INFO] building autoencoder...")
        self.autoencoder = ConvAutoencoder.build(self.img_size[0], self.img_size[1], self.channel)
        opt = Adam(lr=self.INIT_LR, decay=self.INIT_LR / self.EPOCHS)
        self.autoencoder.compile(loss="mse", optimizer=opt)
        # train the convolutional autoencoder
        self.H = self.autoencoder.fit(
            trainX, trainX,
            validation_data=(testX, testX),
            epochs=self.EPOCHS,
            batch_size=self.BS)

        print("[INFO] saving autoencoder...")
        if output_model_path!="":
            self.autoencoder.save(output_model_path, save_format="h5")
        return self.H

    def visualize_predictions(self,decoded, gt, samples=10):
        # initialize our list of output images
        outputs = None
        # loop over our number of output samples
        for i in range(0, samples):
            # grab the original image and reconstructed image
            original = (gt[i] * 255).astype("uint8")
            recon = (decoded[i] * 255).astype("uint8")
            # stack the original and reconstructed image side-by-side
            output = np.hstack([original, recon])
            # if the outputs array is empty, initialize it as the current
            # side-by-side image display
            if outputs is None:
                outputs = output
            # otherwise, vertically stack the outputs
            else:
                outputs = np.vstack([outputs, output])
        # return the output images
        return outputs

    def predict(self,testX,_vis="", _plot=""):
        print("[INFO] making predictions...")
        decoded = self.autoencoder.predict(testX)
        if _vis != "":
            vis = self.visualize_predictions(decoded, testX)
            cv2.imwrite(_vis, vis)
        # construct a plot that plots and saves the training history
        if _plot != "":
            N = np.arange(0, self.EPOCHS)
            plt.style.use("ggplot")
            plt.figure()
            plt.plot(N, self.H.history["loss"], label="train_loss")
            plt.plot(N, self.H.history["val_loss"], label="val_loss")
            plt.title("Training Loss and Accuracy")
            plt.xlabel("Epoch #")
            plt.ylabel("Loss/Accuracy")
            plt.legend(loc="lower left")
            plt.savefig(_plot)
        return decoded

    def predict_image(self,testX,samples=10,_vis="",_plot=""):
        decoded=self.predict(testX,_vis=_vis,_plot=_plot)
        # initialize our list of output images
        outputs = []
        # loop over our number of output samples
        for i in range(0, samples):
            # grab the original image and reconstructed image
            original = (testX[i] * 255).astype("uint8")
            recon = (decoded[i] * 255).astype("uint8")
            # stack the original and reconstructed image side-by-side
            output = [original,recon]
            # if the outputs array is empty, initialize it as the current
            # side-by-side image display
            outputs.append(output)

        # return the output images
        return outputs

    def load_model(self,model_file):
        print("[INFO] loading autoencoder model...")
        self.autoencoder = load_model(model_file)
        return self.autoencoder

    def create_index(self,Ids,Xs,index_file):
        # _model = "output/autoencoder.h5"
        # _index = "output/index.pickle"

        # load the MNIST dataset
        print("[INFO] loading data...")

        # create the encoder model which consists of *just* the encoder
        # portion of the autoencoder
        self.encoder = Model(inputs=self.autoencoder.input,
                        outputs=self.autoencoder.get_layer("encoded").output)
        # quantify the contents of our input images using the encoder
        print("[INFO] encoding images...")
        features = self.encoder.predict(Xs)

        # construct a dictionary that maps the index of the MNIST training
        # image to its corresponding latent-space representation

        data = {"indexes": Ids, "features": features}
        # write the data dictionary to disk
        print("[INFO] saving index...")
        f = open(index_file, "wb")
        f.write(pickle.dumps(data))
        f.close()
        return self.encoder

    def euclidean(self,a, b):
        # compute and return the euclidean distance between two vectors
        return np.linalg.norm(a - b)

    def perform_search(self,queryFeatures, index, maxResults=64):
        # initialize our list of results
        results = []
        # loop over our index
        for i in range(0, len(index["features"])):
            # compute the euclidean distance between our query features
            # and the features for the current image in our index, then
            # update our results list with a 2-tuple consisting of the
            # computed distance and the index of the image
            d = self.euclidean(queryFeatures, index["features"][i])
            results.append((d, i))
        # sort the results and grab the top ones
        results = sorted(results)[:maxResults]
        # return the list of results
        return results

    def search(self,index_file, queryXs, display=True,  Xs=None,display_num=5, max_result=10):
        index = pickle.loads(open(index_file, "rb").read())
        Ids = index["indexes"]

        encoder = Model(inputs=self.autoencoder.input,
                        outputs=self.autoencoder.get_layer("encoded").output)

        features = encoder.predict(queryXs)
        list_results=[]
        count=0
        for idx,f in enumerate(features):
            count+=1
            results = self.perform_search(f, index, maxResults=max_result)
            images = []
            ids=[]
            # loop over the results
            for (d, j) in results:
                # grab the result image, convert it back to the range
                # [0, 255], and then update the images list
                # image = np.dstack([image] * 3)
                if display and count<=display_num:
                    image = (Xs[j] * 255).astype("uint8")
                    images.append(image)
                ids.append(Ids[j])
            if display and count<=display_num:
                # display the query image
                query = (queryXs[idx] * 255).astype("uint8")
                cv2.imshow("Query", query)
                # build a montage from the results and display it
                montage = build_montages(images, (self.img_size[0], self.img_size[1]), (15, 15))[0]
                cv2.imshow("Results", montage)
                cv2.waitKey(0)
            list_results.append([ids,images])
        return list_results





import torch as th

class Normalize(object):

    def __init__(self, mean, std):
        self.mean = th.FloatTensor(mean).view(1, 3, 1, 1)
        self.std = th.FloatTensor(std).view(1, 3, 1, 1)

    def __call__(self, tensor):
        tensor = (tensor - self.mean) / (self.std + 1e-8)
        return tensor

class Preprocessing(object):

    def __init__(self, type):
        self.type = type
        if type == '2d':
            self.norm = Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        elif type == '3d':
            self.norm = Normalize(mean=[110.6, 103.2, 96.3], std=[1.0, 1.0, 1.0])

    def _zero_pad(self, tensor, size):
        n = size - len(tensor) % size
        if n == size:
            return tensor
        else:
            z = th.zeros(n, tensor.shape[1], tensor.shape[2], tensor.shape[3])
            return th.cat((tensor, z), 0)

    def __call__(self, tensor):
        if self.type == '2d':
            tensor = tensor / 255.0
            tensor = self.norm(tensor)
        elif self.type == '3d':
            tensor = self._zero_pad(tensor, 16)
            tensor = self.norm(tensor)
            tensor = tensor.view(-1, 16, 3, 112, 112)
            tensor = tensor.transpose(1, 2)
        return tensor




lines=open(r'C:\Users\dell\Desktop\UMKG\results\conv-model.txt','r',encoding='utf-8')

print(f'Epoch\tLoss')
for l in lines:
    l=l.strip()
    if l.startswith("Epoch:"):
        fs=l.split(" 	Training Loss: ")
        epoch=int(fs[0].split(":")[1].strip())
        loss=float(fs[1].strip())
        print(f'{epoch}\t{loss}')




import sys
import h5py
import os
import json
import numpy as np
from tqdm import tqdm
from mmkfeatures.fusion  import log
from mmkfeatures.fusion .metadataconfigs import *
from mmkfeatures.fusion.integrity_check import *


def read_CSD(resource, destination=None):
    if (resource is None): raise log.error("No resource specified for computational sequence!", error=True)
    if os.path.isfile(resource) is False:
        log.error("%s file not found, please check the path ..." % resource, error=True)
    try:
        h5handle = h5py.File('%s' % resource, 'r')
    except:
        raise log.error("%s resource is not a valid hdf5 computational sequence format ..." % resource, error=True)
    log.success("Computational sequence read from file %s ..." % resource)

    data=dict(h5handle[list(h5handle.keys())[0]]["data"])
    if "rel_data" in h5handle[list(h5handle.keys())[0]]:
        rel_data=dict(h5handle[list(h5handle.keys())[0]]["rel_data"])
    else:
        rel_data=None
    if "dicts" in h5handle[list(h5handle.keys())[0]]:
        dicts= dict(h5handle[list(h5handle.keys())[0]]["dicts"])
    else:
        dicts=None
    metadata=metadata_to_dict(h5handle[list(h5handle.keys())[0]]["metadata"])

    return h5handle, data,rel_data,dicts,metadata

# writing CSD files to disk
def write_CSD(data, rel_data,dicts, metadata, rootName, destination, compression, compression_opts, full_chunk_shape):
    log.status("Writing the <%s> computational sequence data to %s" % (rootName, destination))
    if compression is not None:
        log.advise("Compression with %s and opts -%d" % (compression, compression_opts))
    # opening the file
    writeh5Handle = h5py.File(destination, 'w')
    # creating the root handle
    rootHandle = writeh5Handle.create_group(rootName)

    # writing the data
    dataHandle = rootHandle.create_group("data")
    pbar = log.progress_bar(total=len(data.keys()), unit=" Contents", leave=False)
    for vid in data:
        vidHandle = dataHandle.create_group(vid)
        from mmkfeatures.fusion.mm_features_node import MMFeaturesNode
        import numpy
        temp=MMFeaturesNode("")
        if "compressed" in metadata.keys() and metadata["compressed"].lower()=="true":
            keys=temp.get_all_validated_keys_alias()
        else:
            keys=temp.get_all_validated_keys()
        if compression is not None:
            '''
            vidHandle.create_dataset("features", data=data[vid]["features"], compression=compression,
                                     compression_opts=compression_opts)
            vidHandle.create_dataset("intervals", data=data[vid]["intervals"], compression=compression,
                                     compression_opts=compression_opts)
            '''
            for k in keys:
                if k in data[vid].keys():
                    if k in ["raw","objects"] or k in ['RAW',"OBJ"]:
                        rawHandeler = vidHandle.create_group(k)
                        for idx, raw in enumerate(data[vid][k]):
                            rawHandeler.create_dataset(str(idx), data=raw,
                                                     compression=compression,
                                                     compression_opts=compression_opts
                                                     )
                    elif k in ["features","intervals"] or k in ['F','ITV']:
                        #dt = h5py.vlen_dtype(np.dtype('float64'))
                        #new_data=np.array(data[vid][k])

                        vidHandle.create_dataset(k, data=data[vid][k],
                                                  compression=compression,
                                          compression_opts=compression_opts
                                                 )

                        '''
                        if k=="features" or k=="F":
                            fHandeler = vidHandle.create_group(k)
                            for idx, feature in enumerate(data[vid][k]):
                                fHandeler.create_dataset(str(idx), data=feature,
                                                         compression=compression,
                                                         compression_opts=compression_opts
                                                         )
                        
                        else:
                            vidHandle.create_dataset(k, data=data[vid][k],
                                                     compression=compression,
                                                     compression_opts=compression_opts
                                                    )
                        '''
                    else:
                        # print(k)
                        # print(data[vid][k])
                        if type(data[vid][k])==list:
                            #for x in data[vid][k]:
                            # data=numpy.array(data[vid][k],dtype='S4')
                            # print("list = ",k)
                            vidHandle.create_dataset(k, data=str(data[vid][k]))
                        elif type(data[vid][k])==dict:
                            vidHandle.create_dataset(k,data=str(data[vid][k]))
                        else:
                            vidHandle.create_dataset(k, data=data[vid][k])
        else:
            '''
            vidHandle.create_dataset("features", data=data[vid]["features"])
            vidHandle.create_dataset("intervals", data=data[vid]["intervals"])
            '''
            for k in keys:
                if k in data[vid].keys():
                    vidHandle.create_dataset(k, data=data[vid][k])


        pbar.update(1)
    pbar.close()

    pbar = log.progress_bar(total=len(rel_data.keys()), unit=" Relationships", leave=False)
    relDataHandler = rootHandle.create_group("rel_data")
    for rel_id in rel_data:
        relHandler = relDataHandler.create_group(rel_id)
        keys=rel_data[rel_id].keys()
        for k in keys:
            if type(rel_data[rel_id][k]) == list:
                relHandler.create_dataset(k, data=str(rel_data[rel_id][k]))
            elif type(rel_data[rel_id][k]) == dict:
                relHandler.create_dataset(k, data=str(rel_data[rel_id][k]))
            else:
                relHandler.create_dataset(k, data=rel_data[rel_id][k])
        pbar.update(1)
    pbar.close()

    pbar = log.progress_bar(total=len(rel_data.keys()), unit=" Dictionaries", leave=False)
    dictDataHandler = rootHandle.create_group("dicts")
    for dict_id in dicts:
        dictHandler = dictDataHandler.create_group(dict_id)
        keys = dicts[dict_id].keys()
        for k in keys:
            if type(dicts[dict_id][k]) == list:
                dictHandler.create_dataset(k, data=str(dicts[dict_id][k]))
            elif type(dicts[dict_id][k]) == dict:
                dictHandler.create_dataset(k, data=str(dicts[dict_id][k]))
            else:
                dictHandler.create_dataset(k, data=dicts[dict_id][k])
        pbar.update(1)
    pbar.close()

    log.success("<%s> computational sequence data successfully wrote to %s" % (rootName, destination))
    log.status("Writing the <%s> computational sequence metadata to %s" % (rootName, destination))
    # writing the metadata
    metadataHandle = rootHandle.create_group("metadata")
    for metadataKey in metadata.keys():
        metadataHandle.create_dataset(metadataKey, (1,), dtype=h5py.special_dtype(
            vlen=unicode) if sys.version_info.major is 2 else h5py.special_dtype(vlen=str))
        cast_operator = unicode if sys.version_info.major is 2 else str
        metadataHandle[metadataKey][0] = cast_operator(json.dumps(metadata[metadataKey]))

    writeh5Handle.close()

    log.success("<%s> computational sequence metadata successfully wrote to %s" % (rootName, destination))
    log.success("<%s> computational sequence successfully wrote to %s ..." % (rootName, destination))


def metadata_to_dict(metadata_):
    if (type(metadata_) is dict):
        return metadata_
    else:
        metadata = {}
        for key in metadata_.keys():
            try:
                metadata[key] = json.loads(metadata_[key][0])
            except:
                try:
                    metadata[key] = str(metadata_[key][0])
                except:
                    log.error("Metadata %s is in wrong format. Exiting ...!" % key)
        return metadata


import torch
import torch.nn as nn


class FixedRnn(nn.Module):
    # Custom RNN that matches the implementation provided by Reed et al.
    # We cannot use PyTorch's RNN because Reed's implementation does not use
    # bias for the hidden layer (b_{hh}) during the computation of the first
    # time step.

    def __init__(self, num_steps, emb_dim):
        super().__init__()
        self.i2h = nn.Linear(emb_dim, emb_dim)
        self.h2h = nn.Linear(emb_dim, emb_dim)

        self.num_steps = num_steps
        self.relu = torch.nn.functional.relu

    def forward(self, txt):
        res = []
        for i in range(self.num_steps):
            i2h = self.i2h(txt[:, i]).unsqueeze(1)
            if i == 0:
                output = self.relu(i2h)
            else:
                h2h = self.h2h(res[i - 1])
                output = self.relu(i2h + h2h)
            res.append(output)

        res = torch.cat(res, dim=1)
        res = torch.mean(res, dim=1)
        return res



#CMU Multimodal SDK, CMU Multimodal Model SDK

#Multi-attention Recurrent Network for Human Communication Comprehension, Amir Zadeh, Paul Pu Liang, Soujanya Poria, Erik Cambria, Prateek Vij, Louis-Philippe Morency - https://arxiv.org/pdf/1802.00923.pdf

#in_modalities: is a list of inputs from each modality - the first dimension of all the modality inputs must be the same, it will be the batch size. The second dimension is the feature dimension. There are a total of n modalities.

#attention_model: is a pytorch nn.Sequential which takes in an input with size (bs * m0+...+mn) with m_i being the dimensionality of the features in modality i. Output is the (bs * (m0+...+mn)*num_atts).

#dim_reduce_nets: is a list of pytorch nn.Sequential which takes in an input with size (bs*(mi*num_atts))
#num_atts is the number of attentions

#num_atts: number of attentions


import torch
import time
from torch import nn
import torch.nn.functional as F


class MultipleAttentionFusion(nn.Module):

	def __init__(self,attention_model,dim_reduce_nets,num_atts):
		super(MultipleAttentionFusion, self).__init__()
		self.attention_model=attention_model
		self.dim_reduce_nets=dim_reduce_nets
		self.num_atts=num_atts

	def __call__(self,in_modalities):
		return self.fusion(in_modalities)

	def fusion(self,in_modalities):

		#getting some simple integers out
		num_modalities=len(in_modalities)
		#simply the tensor that goes into attention_model
		in_tensor=torch.cat(in_modalities,dim=1)
		#calculating attentions
		atts=F.softmax(self.attention_model(in_tensor),dim=1)
		#calculating the tensor that will be multiplied with the attention
		out_tensor=torch.cat([in_modalities[i].repeat(1,self.num_atts) for i in range(num_modalities)],dim=1)
		#calculating the attention
		att_out=atts*out_tensor
	
		#now to apply the dim_reduce networks
		#first back to however modalities were in the problem
		start=0
		out_modalities=[]
		for i in range(num_modalities):
			modality_length=in_modalities[i].shape[1]*self.num_atts
			out_modalities.append(att_out[:,start:start+modality_length])
			start=start+modality_length
	
		#apply the dim_reduce
		dim_reduced=[self.dim_reduce_nets[i](out_modalities[i]) for i in range(num_modalities)]
		#multiple attention done :)
		return dim_reduced,out_modalities

	def forward(self, x):
		print("Not yet implemented for nn.Sequential")
		exit(-1)



if __name__=="__main__":
	print("This is a module and hence cannot be called directly ...")
	print("A toy sample will now run ...")
	
	from torch.autograd import Variable
	import torch.nn.functional as F
	import numpy

	inputx=Variable(torch.Tensor(numpy.array(numpy.zeros([32,40]))),requires_grad=True)
	inputy=Variable(torch.Tensor(numpy.array(numpy.zeros([32,12]))),requires_grad=True)
	inputz=Variable(torch.Tensor(numpy.array(numpy.zeros([32,20]))),requires_grad=True)
	modalities=[inputx,inputy,inputz]

	#simple functions for toy example, 4 times extract attentions hence 72*4
	my_attention =	nn.Sequential(nn.Linear(72,72*4))
	small_netx =	nn.Sequential(nn.Linear(160,10))
	small_nety =	nn.Sequential(nn.Linear(48,20))
	small_netz =	nn.Sequential(nn.Linear(80,30))

	smalls_nets=[small_netx,small_nety,small_netz]

	fmodel=MultipleAttentionFusion(my_attention,smalls_nets,4)	
	out=fmodel(modalities)

	print("Output")
	print(out[0])
	print(out)
	print("Toy sample finished ...")






import sys
from datetime import datetime
from colorama import Fore
from tqdm import tqdm

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    OKPURPLE = '\033[0;35m'
    OKADVISORY = '\033[1;36m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def success(msgstring,destination=sys.stdout,verbose=True):
	now=datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
	if type(destination) is not list:
		destination=[destination]
	if verbose==False:
		return
	for dest in destination:
		print(bcolors.OKGREEN+bcolors.BOLD+"[%s] | Success | "%now+ bcolors.ENDC+msgstring,file=dest)

def status(msgstring,destination=sys.stdout,verbose=True,end=None,require_input=False):

	now=datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

	if type(destination) is not list:
		destination=[destination]

	if verbose==False:
		return

	input_from_user=None

	for dest in destination:
		if end is None:
			if require_input:
				if dest==sys.stdout:
					inp_f=raw_input if sys.version_info[0]<3 else input
					input_from_user=inp_f(bcolors.OKBLUE +bcolors.BOLD+"[%s] | Input   | "%now+bcolors.ENDC + msgstring)
				else:
					print (bcolors.OKBLUE +bcolors.BOLD+"[%s] | Status  | "%now+bcolors.ENDC + msgstring,file=dest)
					
			else:
				print (bcolors.OKBLUE +bcolors.BOLD+"[%s] | Status  | "%now+bcolors.ENDC + msgstring,file=dest)
		else:
			print (bcolors.OKBLUE +bcolors.BOLD+"[%s] | Status  | "%now+bcolors.ENDC + msgstring,file=dest,end="\r")

	if input_from_user!=None:
		return input_from_user

def advisory(msgstring,destination=sys.stdout,verbose=True):
	now=datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

	if type(destination) is not list:
		destination=[destination]

	if verbose==False:
		return

	for dest in destination:
		print (bcolors.OKADVISORY +bcolors.BOLD+"[%s] | Advise  | "%now+bcolors.ENDC + msgstring,file=dest)

advise=advisory

def progress_bar(total,data=None,unit="iters",postfix="",leave=False):
	if data is None:
		return tqdm(total=total , postfix=postfix,unit=unit, leave=leave)
	#TQDM has issue with the formatting and {bar}
		#return tqdm(total=total , postfix=postfix,unit=unit, leave=leave,bar_format="%s{l_bar}%s{bar}%s{r_bar}%s" % (Fore.YELLOW,Fore.GREEN,Fore.YELLOW,Fore.RESET))
	else:
		return tqdm(data, total=total , postfix=postfix,unit=unit, leave=leave)
		#return tqdm(data, total=total , postfix=postfix,unit=unit, leave=leave,bar_format="%s{l_bar}%s{bar}%s{r_bar}%s" % (Fore.YELLOW,Fore.GREEN,Fore.YELLOW,Fore.RESET))

def error(msgstring,error=False,errorType=RuntimeError,destination=sys.stdout,verbose=True):
	now=datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

	if type(destination) is not list:
		destination=[destination]

	if verbose==False:
		if error:
			raise errorType(msgstring)
		else:
			return

	if error:
		for dest in destination:
			print (bcolors.FAIL +bcolors.BOLD+"[%s] | Error   | "%now+bcolors.ENDC + msgstring,file=dest)
		raise errorType(msgstring)
	else:
		for dest in destination:
			print (bcolors.WARNING +bcolors.BOLD+"[%s] | Warning | "%now+bcolors.ENDC + msgstring,file=dest)

def warning(msgstring,destination=sys.stdout,verbose=True):
	error(msgstring=msgstring,destination=destination,verbose=verbose)

def progress_spinner(message,progress,speed=1./5000):
	speed=float(speed)
	status ("%s%s"%(message,'/-\|'[int(progress*speed)%4]),end="\r")

spinner=progress_spinner





import os
from typing import List, Tuple

import cv2
import numpy as np
from skimage.feature import local_binary_pattern

from mmkfeatures.image.extractor import Extractor


class LBPExtractor(Extractor):
    
    def __init__(self, base_route: str, points: int, radius: int, size: int, grid_x: int, grid_y: int,
                 batch_size=128, method: str = 'uniform', color_code: int = cv2.COLOR_BGR2GRAY):
        
        super().__init__(base_route=base_route, size=size, batch_size=batch_size)
        
        self.points: int = points
        self.radius: int = radius
        self.method: str = method
        self.color_code: int = color_code
        
        self.bins: int = 256
        self.range: Tuple[int, int] = (0, 255)
        print(self.height,grid_y)
        self.height_step =  self.height // grid_y
        self.width_step  = self.width // grid_x
    
    def extract(self, image_route: str) -> np.ndarray:
        """
        Generates an histogram for the given image route.
        
        :param image_route:
        :return:
        """
        
        image = self._read_image(image_route)
        lbp = self._extract_lbp(image)
        
        histogram_grid = self._convert_lbp_to_histogram(lbp)
        
        return np.array(histogram_grid).flatten()
    
    def _read_image(self, image_route: str) -> np.ndarray:
        image = cv2.imread(image_route)
        gray_image = cv2.cvtColor(image, self.color_code)
        
        return gray_image
    
    def _extract_lbp(self, image: np.ndarray) -> np.ndarray:
        return local_binary_pattern(image, self.points, self.radius, method=self.method)
    
    def _convert_lbp_to_histogram(self, image: np.ndarray) -> List[np.ndarray]:
        grid = []
        
        for y in range(0, self.height, self.height_step):
            for x in range(0, self.width, self.width_step):
                y_limit = y + self.height_step
                x_limit = x + self.width_step
                
                batch = image[y:y_limit, x:x_limit]
                
                grid.append(self._convert_image_chunk_to_histogram(batch))
        
        return grid
    
    def _convert_image_chunk_to_histogram(self, lbp_chunk: np.ndarray) -> np.ndarray:
        hist, _ = np.histogram(lbp_chunk, bins=self.bins, range=self.range)
        
        return hist
    
    def _find_features_size(self) -> int:
        example_image_route = os.path.join(self.base_route, self.directory_iterator.filenames[0])
        return len(self.extract(image_route=example_image_route))



import sys
import os
from numpy import expand_dims
from numpy import zeros
from numpy import ones
from numpy.random import randn
from numpy.random import randint
from keras.datasets.fashion_mnist import load_data
from keras import optimizers
from keras.models import Model
from keras.layers import Input
from keras.layers import Dense
from keras.layers import Reshape
from keras.layers import Flatten
from keras.layers import Conv2D
from keras.layers import Conv2DTranspose
from keras.layers import LeakyReLU
from keras.layers import Dropout
from keras.layers import Embedding
from keras.layers import Concatenate
from keras.models import load_model
from matplotlib import pyplot
from numpy import asarray
import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
import h5py
import cv2
sys.path.append("..")
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
physical_devices = tf.config.list_physical_devices('GPU')
tf.config.experimental.set_memory_growth(physical_devices[0], True)

class CGanNode:

	def __init__(self,name,root_path=""):
		self.root_path=f"{root_path}/{name}"
		if not os.path.exists(self.root_path):
			os.mkdir(self.root_path)
		self.model_path=f"{self.root_path}/{name}.h5"
		self.data_path=f"{self.root_path}/data.h5"
		self.metrics=[]


	def get_files(self,file_dir):

		labels=os.listdir(file_dir)

		files_set = []
		lbs_set = []
		for idx, label in enumerate(labels):
			files = []
			lbs = []
			for file in os.listdir(file_dir + '/' + label):
				files.append(file_dir + '/' + label + '/' + file)
				lbs.append(idx)  # 添加标签，该类标签为0，此为2分类例子，多类别识别问题自行添加
			files_set.append(files)
			lbs_set.append(lbs)

		files_tuple = ()
		lbs_tuple = ()
		for idx, label in enumerate(labels):
			files_tuple = files_tuple + tuple(files_set[idx])
			lbs_tuple = lbs_tuple + tuple(lbs_set[idx])

		# 把cat和dog合起来组成一个list（img和lab）
		image_list = np.hstack(files_tuple)
		label_list = np.hstack(lbs_tuple)

		# 利用shuffle打乱顺序
		temp = np.array([image_list, label_list])
		temp = temp.transpose()
		np.random.shuffle(temp)

		# 从打乱的temp中再取出list（img和lab）
		image_list = list(temp[:, 0])
		label_list = list(temp[:, 1])
		label_list = [int(i) for i in label_list]

		return image_list, label_list

	def create_datasets(self,train_dir=""):
		if train_dir=="":
			train_dir=self.data_path+"/data"
		# train_dir = 'datasets/Medical_MNIST'
		image_list, label_list = self.get_files(train_dir)

		# print(len(image_list))
		# print(len(label_list))

		# 450为数据长度的20%
		percent20 = int(len(image_list) * 0.2)
		dim = 28
		Train_image = np.random.rand(len(image_list) - percent20, dim, dim).astype('float32')
		Train_label = np.random.rand(len(image_list) - percent20, 1).astype('float32')

		Test_image = np.random.rand(percent20, dim, dim).astype('float32')
		Test_label = np.random.rand(percent20, 1).astype('float32')

		for i in range(len(image_list) - percent20):
			img = plt.imread(image_list[i])
			# img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
			shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
			Train_image[i] = np.array(shrink_img)
			Train_label[i] = np.array(label_list[i])

		for i in range(len(image_list) - percent20, len(image_list)):
			img = plt.imread(image_list[i])
			# img_gray=cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
			shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
			Test_image[i + percent20 - len(image_list)] = np.array(shrink_img)
			Test_label[i + percent20 - len(image_list)] = np.array(label_list[i])

		# Create a new file
		f = h5py.File(self.data_path, 'w')
		f.create_dataset('X_train', data=Train_image)
		f.create_dataset('y_train', data=Train_label)
		f.create_dataset('X_test', data=Test_image)
		f.create_dataset('y_test', data=Test_label)
		f.close()

	# define the standalone discriminator model
	def define_discriminator(self,in_shape=(28,28,1), n_classes=10):
		# label input
		in_label = Input(shape=(1,))
		# embedding for categorical input
		li = Embedding(n_classes, 50)(in_label)
		# scale up to image dimensions with linear activation
		n_nodes = in_shape[0] * in_shape[1]
		li = Dense(n_nodes)(li)
		# reshape to additional channel
		li = Reshape((in_shape[0], in_shape[1], 1))(li)
		# image input
		in_image = Input(shape=in_shape)
		# concat label as a channel
		merge = Concatenate()([in_image, li])
		# downsample
		fe = Conv2D(128, (3,3), strides=(2,2), padding='same')(merge)
		fe = LeakyReLU(alpha=0.2)(fe)
		# downsample
		fe = Conv2D(128, (3,3), strides=(2,2), padding='same')(fe)
		fe = LeakyReLU(alpha=0.2)(fe)
		# flatten feature maps
		fe = Flatten()(fe)
		# dropout
		fe = Dropout(0.4)(fe)
		# output
		out_layer = Dense(1, activation='sigmoid')(fe)
		# define model
		model = Model([in_image, in_label], out_layer)
		# compile model
		opt = optimizers.adam_v2.Adam(lr=0.0002, beta_1=0.5)
		model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
		return model

	# define the standalone generator model
	def define_generator(self,latent_dim, n_classes=10):
		# label input
		in_label = Input(shape=(1,))
		# embedding for categorical input
		li = Embedding(n_classes, 50)(in_label)
		# linear multiplication
		n_nodes = 7 * 7
		li = Dense(n_nodes)(li)
		# reshape to additional channel
		li = Reshape((7, 7, 1))(li)
		# image generator input
		in_lat = Input(shape=(latent_dim,))
		# foundation for 7x7 image
		n_nodes = 128 * 7 * 7
		gen = Dense(n_nodes)(in_lat)
		gen = LeakyReLU(alpha=0.2)(gen)
		gen = Reshape((7, 7, 128))(gen)
		# merge image gen and label input
		merge = Concatenate()([gen, li])
		# upsample to 14x14
		gen = Conv2DTranspose(128, (4,4), strides=(2,2), padding='same')(merge)
		gen = LeakyReLU(alpha=0.2)(gen)
		# upsample to 28x28
		gen = Conv2DTranspose(128, (4,4), strides=(2,2), padding='same')(gen)
		gen = LeakyReLU(alpha=0.2)(gen)
		# output
		out_layer = Conv2D(1, (7,7), activation='tanh', padding='same')(gen)
		# define model
		model = Model([in_lat, in_label], out_layer)
		return model

	# define the combined generator and discriminator model, for updating the generator
	def define_gan(self,g_model, d_model):
		# make weights in the discriminator not trainable
		d_model.trainable = False
		# get noise and label inputs from generator model
		gen_noise, gen_label = g_model.input
		# get image output from the generator model
		gen_output = g_model.output
		# connect image output and label input from generator as inputs to discriminator
		gan_output = d_model([gen_output, gen_label])
		# define gan model as taking noise and label and outputting a classification
		model = Model([gen_noise, gen_label], gan_output)
		# compile model
		opt = optimizers.adam_v2.Adam(lr=0.0002, beta_1=0.5)
		model.compile(loss='binary_crossentropy', optimizer=opt)
		return model

	# load fashion mnist images
	def load_real_samples(self):
		# load dataset
		(trainX, trainy), (_, _) = load_data()
		# expand to 3d, e.g. add channels
		X = expand_dims(trainX, axis=-1)
		# convert from ints to floats
		X = X.astype('float32')
		# scale from [0,255] to [-1,1]
		X = (X - 127.5) / 127.5
		return [X, trainy]

	def load_my_samples(self,data_path):
		# load dataset
		import h5py
		import numpy as np
		# (trainX, _), (_, _) = load_data()
		train_dataset = h5py.File(data_path, 'r')
		train_set_x_orig = np.array(train_dataset['X_train'][:])  # your train set features
		train_set_y_orig = np.array(train_dataset['y_train'][:])  # your train set labels
		test_set_x_orig = np.array(train_dataset['X_test'][:])  # your train set features
		test_set_y_orig = np.array(train_dataset['y_test'][:])  # your train set labels
		train_dataset.close()

		# print(train_set_x_orig.shape)
		#print(train_set_y_orig.shape)

		#print(train_set_x_orig.max())
		#print(train_set_x_orig.min())

		#print(test_set_x_orig.shape)
		#print(test_set_y_orig.shape)
		X=train_set_x_orig
		# expand to 3d, e.g. add channels
		X = expand_dims(train_set_x_orig, axis=-1)
		# print("expand dims")
		# print(X.shape)
		# convert from ints to floats
		X = X.astype('float32')
		# scale from [0,255] to [-1,1]
		X = (X - 127.5) / 127.5
		return [X, train_set_y_orig]

	# # select real samples
	def generate_real_samples(self,dataset, n_samples):
		# split into images and labels
		images, labels = dataset
		# choose random instances
		ix = randint(0, images.shape[0], n_samples)
		# select images and labels
		X, labels = images[ix], labels[ix]
		# generate class labels
		y = ones((n_samples, 1))
		return [X, labels], y

	# generate points in latent space as input for the generator
	def generate_latent_points(self,latent_dim, n_samples, n_classes=10):
		# generate points in the latent space
		x_input = randn(latent_dim * n_samples)
		# reshape into a batch of inputs for the network
		z_input = x_input.reshape(n_samples, latent_dim)
		# generate labels
		labels = randint(0, n_classes, n_samples)
		return [z_input, labels]

	# use the generator to generate n fake examples, with class labels
	def generate_fake_samples(self,generator, latent_dim, n_samples):
		# generate points in latent space
		z_input, labels_input = self.generate_latent_points(latent_dim, n_samples)
		# predict outputs
		images = generator.predict([z_input, labels_input])
		# create class labels
		y = zeros((n_samples, 1))
		return [images, labels_input], y

	# train the generator and discriminator
	def train(self,g_model, d_model, gan_model, dataset, latent_dim, n_epochs=100, n_batch=128,model_path=""):
		bat_per_epo = int(dataset[0].shape[0] / n_batch)
		half_batch = int(n_batch / 2)
		# manually enumerate epochs
		for i in range(n_epochs):
			# enumerate batches over the training set
			for j in range(bat_per_epo):
				# get randomly selected 'real' samples
				[X_real, labels_real], y_real = self.generate_real_samples(dataset, half_batch)
				# update discriminator model weights
				d_loss1, _ = d_model.train_on_batch([X_real, labels_real], y_real)
				# generate 'fake' examples
				[X_fake, labels], y_fake = self.generate_fake_samples(g_model, latent_dim, half_batch)
				# update discriminator model weights
				d_loss2, _ = d_model.train_on_batch([X_fake, labels], y_fake)
				# prepare points in latent space as input for the generator
				[z_input, labels_input] = self.generate_latent_points(latent_dim, n_batch)
				# create inverted labels for the fake samples
				y_gan = ones((n_batch, 1))
				# update the generator via the discriminator's error
				g_loss = gan_model.train_on_batch([z_input, labels_input], y_gan)
				# summarize loss on this batch
				print('>%d, %d/%d, d1=%.3f, d2=%.3f g=%.3f' %
					(i+1, j+1, bat_per_epo, d_loss1, d_loss2, g_loss))
				self.metrics.append([str(i+1),str(j+1),str(bat_per_epo),str(d_loss1),str(d_loss2),str(g_loss)])

		# save the generator model
		g_model.save(model_path)

	def export_performance(self,export_file=""):
		if export_file=="":
			export_file=self.root_path+"/performance.csv"
		f_out=open(export_file,"w",encoding='utf-8')
		for metric in self.metrics:
			line="\t".join(metric)
			f_out.write(line+"\n")
		f_out.close()

	def build(self,n_epochs=10):
		# size of the latent space
		latent_dim = 100
		# create the discriminator
		d_model = self.define_discriminator()
		# create the generator
		g_model = self.define_generator(latent_dim)
		# create the gan
		gan_model = self.define_gan(g_model, d_model)
		# load image data
		dataset = self.load_my_samples(data_path=self.data_path)
		# train model
		self.train(g_model, d_model, gan_model, dataset, latent_dim,model_path=self.model_path,n_epochs=n_epochs)

	# create and save a plot of generated images
	def save_plot(self,examples, n):
		# plot images
		for i in range(n * n):
			# define subplot
			pyplot.subplot(n, n, 1 + i)
			# turn off axis
			pyplot.axis('off')
			# plot raw pixel data
			pyplot.imshow(examples[i, :, :, 0], cmap='gray_r')
		pyplot.show()

	def generate(self,n_samples=10):
		# load model
		model = load_model(self.model_path)
		# generate images
		latent_points, labels = self.generate_latent_points(100, 100)
		# specify labels
		labels = asarray([x for _ in range(n_samples) for x in range(10)])
		# generate images
		X = model.predict([latent_points, labels])
		# scale from [-1,1] to [0,1]
		X = (X + 1) / 2.0
		# plot the result
		self.save_plot(X, n_samples)



# import the necessary packages
import numpy as np
import cv2
import imutils
class ColorDescriptor:
	def __init__(self, bins):
		# store the number of bins for the 3D histogram
		self.bins = bins
	def describe(self, image):
		# convert the image to the HSV color space and initialize
		# the features used to quantify the image
		image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
		features = []
		# grab the dimensions and compute the center of the image
		(h, w) = image.shape[:2]
		(cX, cY) = (int(w * 0.5), int(h * 0.5))
		# divide the image into four rectangles/segments (top-left,
		# top-right, bottom-right, bottom-left)
		segments = [(0, cX, 0, cY), (cX, w, 0, cY), (cX, w, cY, h),
			(0, cX, cY, h)]
		# construct an elliptical mask representing the center of the
		# image
		(axesX, axesY) = (int(w * 0.75) // 2, int(h * 0.75) // 2)
		ellipMask = np.zeros(image.shape[:2], dtype = "uint8")
		cv2.ellipse(ellipMask, (cX, cY), (axesX, axesY), 0, 0, 360, 255, -1)
		# loop over the segments
		for (startX, endX, startY, endY) in segments:
			# construct a mask for each corner of the image, subtracting
			# the elliptical center from it
			cornerMask = np.zeros(image.shape[:2], dtype = "uint8")
			cv2.rectangle(cornerMask, (startX, startY), (endX, endY), 255, -1)
			cornerMask = cv2.subtract(cornerMask, ellipMask)
			# extract a color histogram from the image, then update the
			# feature vector
			hist = self.histogram(image, cornerMask)
			features.extend(hist)
		# extract a color histogram from the elliptical region and
		# update the feature vector
		hist = self.histogram(image, ellipMask)
		features.extend(hist)
		# return the feature vector
		return features

	def histogram(self, image, mask):
		# extract a 3D color histogram from the masked region of the
		# image, using the supplied number of bins per channel
		hist = cv2.calcHist([image], [0, 1, 2], mask, self.bins,
			[0, 180, 0, 256, 0, 256])
		# normalize the histogram if we are using OpenCV 2.4
		if imutils.is_cv2():
			hist = cv2.normalize(hist).flatten()
		# otherwise handle for OpenCV 3+
		else:
			hist = cv2.normalize(hist, hist).flatten()
		# return the histogram
		return hist



import os
from mmkfeatures.speech.MFCC import *

import scipy.io.wavfile as wav
import mmkfeatures.speech.feature_extraction_functions
import librosa
import librosa.display
import matplotlib.pyplot as plt
import sklearn
import wave
import pylab as pl
import numpy as np
import mmkfeatures.speech.Volume as vp
from mmkfeatures.speech.ZeroCR import *

class AudioFeaturesWrapper:
    def __init__(self):
        pass

    def get_mfcc_features(self,sound_file):
        # "../sounds/english.wav"
        (rate, sig) = wav.read(sound_file)
        mfcc_feat = mfcc(sig, rate)
        d_mfcc_feat = delta(mfcc_feat, 2)
        fbank_feat = logfbank(sig, rate)
        print(fbank_feat[1:3, :])
        return fbank_feat[1:3, :]

    def get_mfcc_features2(self,sound_file,n_mfcc=24):
        sig, rate = librosa.load(sound_file)
        # (rate, sig) = wav.read(sound_file)
        # sig = sig[0:int(3.5 * rate)]
        features=mmkfeatures.speech.feature_extraction_functions.get_feature(rate,sig,n_mfcc=n_mfcc)
        return features

    def show_features(self,sound_file):
        x, fs = librosa.load(sound_file)
        # librosa.display.waveplot(x, sr=fs)
        mfccs = librosa.feature.mfcc(x, sr=fs)
        librosa.display.specshow(mfccs, sr=fs, x_axis='time')
        plt.show()

    def show_plot(self,sound_file):
        x, fs = librosa.load(sound_file)
        librosa.display.waveplot(x, sr=fs)
        plt.show()

    def show_scaled_features(self,sound_file):
        x, fs = librosa.load(sound_file)
        # librosa.display.waveplot(x, sr=fs)
        mfccs = librosa.feature.mfcc(x, sr=fs)
        mfccs = sklearn.preprocessing.scale(mfccs, axis=1)
        print(mfccs.mean(axis=1))
        print(mfccs.var(axis=1))
        librosa.display.specshow(mfccs, sr=fs, x_axis='time')
        plt.show()

    def get_volume_features(self,sound_file):
        # read wave file and get parameters.
        fw = wave.open(sound_file, 'rb')
        params = fw.getparams()
        print(params)
        nchannels, sampwidth, framerate, nframes = params[:4]
        strData = fw.readframes(nframes)
        waveData = np.fromstring(strData, dtype=np.int16)
        waveData = waveData * 1.0 / max(abs(waveData))  # normalization
        fw.close()

        # calculate volume
        frameSize = 256
        overLap = 128
        volume11 = vp.calVolume(waveData, frameSize, overLap)
        volume12 = vp.calVolumeDB(waveData, frameSize, overLap)
        return [waveData,volume11,volume12] # amplitude, absSum, Decibel(dB)

    def get_zero_cr_features(self,sound_file):
        # read wave file and get parameters.
        fw = wave.open(sound_file, 'rb')
        params = fw.getparams()
        print(params)
        nchannels, sampwidth, framerate, nframes = params[:4]
        strData = fw.readframes(nframes)
        waveData = np.fromstring(strData, dtype=np.int16)
        waveData = waveData * 1.0 / max(abs(waveData))  # normalization
        fw.close()

        # calculate Zero Cross Rate
        frameSize = 256
        overLap = 0
        zcr = ZeroCR(waveData, frameSize, overLap)
        return zcr

if __name__ == '__main__':
    audio_feat_wrapper=AudioFeaturesWrapper()
    sound_file="../dataset/sounds/english.wav"
    # features=audio_feat_wrapper.get_mfcc_features(sound_file)
    # print(features)
    # print("-=======================")
    features1=audio_feat_wrapper.get_mfcc_features2(sound_file)
    print(features1)
    # audio_feat_wrapper.show_plot(sound_file)
    # audio_feat_wrapper.show_features(sound_file)
    # audio_feat_wrapper.show_scaled_features(sound_file)

    # volume
    # fs_volume=audio_feat_wrapper.get_volume_features(sound_file)
    # print(fs_volume[2])
    # ZeroCR
    fs_zerocr=audio_feat_wrapper.get_volume_features(sound_file)
    print(fs_zerocr)





import numpy as np
from mmkfeatures.fusion.computational_sequence_x import computational_sequence_x
from mmkfeatures.fusion.mm_features_node import MMFeaturesNode
import h5py
import nexusformat.nexus as nx
# import ast
from tqdm import tqdm
import pickle
import cv2
from quickcsv import *

class MMFeaturesLib:
    def __init__(self,root_name="",dataset_name="",file_path="",description="",version="1.0",creator="mmkit-features",contact="",enable_key_compressed=False):
        if file_path=="" or file_path==None:
            self.root_name = root_name
            self.name = dataset_name
            self.meta_dict = {
                "root name": self.root_name,  # name of the featureset
                "computational sequence description": description,  # name of the featureset
                "dimension names": None,
                "computational sequence version": "1.0",  # the version of featureset
                "alignment compatible": "",  # name of the featureset
                "dataset name": self.name,  # featureset belongs to which dataset
                "dataset version": version,  # the version of dataset
                "creator": creator,  # the author of the featureset
                "contact": contact,  # the contact of the featureset
                "featureset bib citation": "",  # citation for the paper related to this featureset
                "dataset bib citation": "",  # citation for the dataset,
                "attributes_alias":"",
                "attributes":""
            }
            self.compseq = computational_sequence_x(self.name)
            self.compseq_data = {}
            self.compseq_rel_data={}
            self.compseq_dicts={}
            self.feature_dimension_num = 1
        else:
            self.compseq = computational_sequence_x(file_path)
            self.meta_dict=self.compseq.metadata
            self.root_name=self.meta_dict["root name"]
            self.name=self.meta_dict["dataset name"]
            self.compseq_data=self.compseq.data
            self.compseq_rel_data=self.compseq.rel_data
            self.compseq_dicts={}
            self.file_path=file_path
        self.use_alias_names=enable_key_compressed
        if self.use_alias_names:
            self.meta_dict["compressed"]="True"
            temp = MMFeaturesNode("")
            self.meta_dict["attributes"] = ','.join(temp.get_all_validated_keys())
            self.meta_dict["attributes_alias"] = ','.join(temp.get_all_validated_keys_alias())
        else:
            self.meta_dict["compressed"]="False"
            self.meta_dict["attributes"] = ""
            self.meta_dict["attributes_alias"] = ""

    def use_alias_name(self):
        return self.use_alias_names

    def set_compressed(self,enable=True):
        temp=MMFeaturesNode("")
        self.meta_dict["attributes"]=','.join(temp.get_all_validated_keys())
        self.meta_dict["attributes_alias"]=','.join(temp.get_all_validated_keys_alias())
        self.meta_dict["compressed"]=str(enable)
        self.use_alias_names=enable

    def set_features_name(self,feature_names):
        self.feature_names=feature_names
        self.feature_dimension_num=len(feature_names)
        self.meta_dict["dimension names"]=feature_names

    '''
    def set_data(self,features):
        self.compseq_data = {}
        ids=features.keys()
        for key in ids:
            self.compseq_data[key] = {}
            self.compseq_data[key]["features"] = np.array(features[key][0])
            # let's assume each video is one minute, hence 60 seconds.
            self.compseq_data[key]["intervals"] = np.array(features[key][1])
        self.compseq.setData(self.compseq_data, self.name)
    '''

    def compress_content(self,content):
        new_content=MMFeaturesNode(content.get_content_id())
        ks=content.keys()
        ks_ori=content.get_all_validated_keys()
        mapped_ks=content.get_all_validated_keys_alias()
        for k in ks:
            new_k=mapped_ks[ks_ori.index(k)]
            new_content.set_item(new_k,content[k])
        new_content.pop("content_id")
        return new_content

    def set_rel_data(self,rel_data):
        self.compseq_rel_data=rel_data
        self.compseq.setRelData(self.compseq_rel_data)

    def set_dicts(self,dicts):
        self.compseq_dicts=dicts
        self.compseq.setDicts(self.compseq_dicts)

    def set_data(self,list_content):

        if self.use_alias_names:
            self.compseq_data = {}
            ks=list(list_content.keys())
            for k in ks:
                self.compseq_data[k]=self.compress_content(list_content[k])
        else:
            self.compseq_data = list_content

        # self.compseq_data = list_content
        self.compseq.setData(self.compseq_data, self.name)

    def get_data(self):
        return self.compseq

    def save_data(self,f_path=None):
        if f_path==None:
            f_path=self.name+".csd"
        self.compseq.deploy(f_path, my_metadata=self.meta_dict)

    def show_structure(self,f_path=None):
        if f_path == None:
            file_name = self.file_path
        else:
            file_name = f_path
        f = nx.nxload(file_name)
        print(f.tree)
        print()

    def show_sample_data_compressed(self,group,max_num):
        meta_keys = list(group["metadata"])
        metadata = group["metadata"]
        print("=====================Start Meta Data======================")
        for mk in meta_keys:
            print(f"{mk} -> {group['metadata'][mk][0]}")
        print("=====================End Meta Data====================")
        print("=====================Start Content Data====================")
        content_ids = list(group["data"])
        for k in content_ids[:max_num]:  # each video
            print("ID: ", k)
            print(list(group["data"][k]))
            content_keys=list(group["data"][k])
            part = group["data"][k]
            if 'F' in content_keys:
                features = part['F']
            else:
                features=[]
            if 'ITV' in content_keys:
                intervals = part['ITV']
            else:
                intervals=[]
            print("[features]: ")
            for f in features[:5]:
                print("\t", f)
            print("number of content features: ", len(features))
            print("[intervals]: ")
            for inter in intervals[:5]:
                print("\t", inter)
            print("number of content intervals: ", len(intervals))
            for kk in part.keys():
                if kk not in ["F", "ITV"]:
                    # print(kk)
                    print(kk, ":  ", part[kk][()])
                    if kk in ["LB", "FID", "LBN"]:
                        labels = part[kk]
                        # print("type of labels",type(labels.value))
                        # print("type of label list", type(labels_list))
                        # print(part[kk].value)
                        for item in eval(part[kk][()]):
                            print("\t\t--> ", item)
                    if kk in ['AT']:
                        for k in eval(part[kk][()]).keys():
                            print("\t\t-->  ", k, " = ", eval(part[kk][()])[k])
        print()
        print("Printing...relationship table")
        rel_ids = list(group["rel_data"])
        for rel_id in rel_ids[:max_num]:
            rel_obj = group["rel_data"][rel_id]
            print(rel_obj["start"][()], "-->", rel_obj["end"][()])

    def show_sample_data_not_compressed(self,group,max_num):
        meta_keys = list(group["metadata"])
        metadata = group["metadata"]
        print("=====================Start Meta Data======================")
        for mk in meta_keys:
            print(f"{mk} -> {group['metadata'][mk][0]}")
        print("=====================End Meta Data====================")
        if "dicts" in group.keys():
            dict_data=group["dicts"]
            for d in dict_data:
                print(d)
                dd=dict_data[d]
                count=0
                for d1 in dd:
                    count+=1
                    print("  "+str(d1)+"  --> ",dd[d1][()])
                    if count>=5:
                        break
                print("  ...")


        print("=====================Start Content Data====================")
        content_ids = list(group["data"])
        for k in content_ids[:max_num]:  # each video
            print("[Content ID]: ", k)
            part = group["data"][k]
            keys=list(group["data"][k])
            if 'features' in keys:
                features = part['features']
            else:
                features=[]
            if 'intervals' in keys:
                intervals = part['intervals']
            else:
                intervals=[]
            print("[features]: ")
            for f in features[:5]:
                print("\t", f)
            print("number of content features: ", len(features))
            print("[intervals]: ")
            for inter in intervals[:5]:
                print("\t", inter)
            print("number of content intervals: ", len(intervals))
            for kk in part.keys():
                if kk not in ["features", "intervals"]:
                    # print(kk)
                    print(kk, ":  ", part[kk][()])
                    if kk in ["labels", "feature_ids", "label_names"]:
                        labels = part[kk]
                        # print("type of labels",type(labels.value))
                        # print("type of label list", type(labels_list))
                        # print(part[kk].value)
                        for item in eval(part[kk][()]):
                            print("\t\t--> ", item)
                    if kk in ['attributes']:
                        print(part[kk][()])
                        for k in eval(part[kk][()]).keys():
                            print("k = ",k)
                            print("\t\t-->  ", k, " = ", eval(part[kk][()])[k])
        print()
        print("Printing...relationship table")
        rel_ids = list(group["rel_data"])
        for rel_id in rel_ids[:max_num]:
            rel_obj=group["rel_data"][rel_id]
            print(rel_obj["start"][()], "-->", rel_obj["end"][()])

    def show_sample_data(self,f_path=None,max_num=5):
        if f_path==None:
            file_name=self.file_path
        else:
            file_name=f_path
        print()
        print("MMF FileName: ",file_name)
        with h5py.File(file_name, "r") as f:
            group = f[self.root_name]
            if 'compressed' in list(group["metadata"]):
                compressed=eval(str(group["metadata"]["compressed"][0]))
            else:
                compressed="false"
            # str(metadata_[key][0])
            is_compressed=False
            if compressed == "True" or compressed == "true":
                is_compressed=True

            if not is_compressed:
                self.show_sample_data_not_compressed(group,max_num)
            else:
                self.show_sample_data_compressed(group,max_num)

    def to_index_file(self,field,index_file_path,index_type="brutal_force"):

        data=self.get_data()
        if index_type=="brutal_force":
            list_id_value = []
            for key in tqdm(data.keys()):
                content = data[key]
                if field in content.keys():
                    v = content[field][()]
                else:
                    v=""
                # print(key,v)
                list_id_value.append([key, v])
            pickle.dump(list_id_value, open(index_file_path, 'wb'))
        if index_type=="inverted_index":
            from mmkfeatures.index.index_inverted_op import InvertedIndex
            inverted_index = InvertedIndex()
            list_id_value = []
            for key in tqdm(data.keys()):
                content = data[key]
                if field in content.keys():
                    v = content[field][()]
                else:
                    v = ""
                list_id_value.append([key, v])
            inverted_index.create_with_data(list_id_value,index_file_path)
        if index_type=="positional_index":
            from mmkfeatures.index.index_positional_op import PositionalIndex
            pos_index = PositionalIndex()
            list_id_value = []
            for key in tqdm(data.keys()):
                content = data[key]
                if field in content.keys():
                    v = content[field][()]
                else:
                    v = ""
                list_id_value.append([key, v])
            pos_index.create_with_data(list_id_value,index_file_path)

    def search_index(self,index_file_path,query,search_type="brutal_force"):
        list_ids = []
        if search_type=="brutal_force":
            list_id_text = pickle.load(open(index_file_path, 'rb'))
            for item in list_id_text:
                text = str(item[1])
                if query in text:
                    list_ids.append(item[0])
            return list_ids
        if search_type=="inverted_index":
            from mmkfeatures.index.index_inverted_op import InvertedIndex
            inverted_index=InvertedIndex()
            return list(inverted_index.search(index_file_path,query))
        if search_type=="positional_index":
            from mmkfeatures.index.index_positional_op import PositionalIndex
            pos_index=PositionalIndex()
            ws=query.split(" ")
            return pos_index.positional_indexing_search_2words(index_file_path,ws[0],ws[1])
        return None

    def to_obj_index(self,index_file,obj_field="objects",index_type="color_descriptor"):
        print("Creating index....")
        if index_type=="color_descriptor":
            from mmkfeatures.image.color_descriptor import ColorDescriptor
            cd = ColorDescriptor((8, 12, 3))
            data = self.get_data()
            f_out = open(index_file, "w")
            for cid in tqdm(data.keys()):
                item = data[cid]
                imgs = item[obj_field]
                for img_id in imgs:
                    # print(image)
                    img = imgs[img_id][()]
                    # print(img)
                    # print(type(img))
                    features = cd.describe(img)
                    features = [str(f) for f in features]
                    # print(feature)
                    feature_str = cid + "," + ",".join(features)
                    f_out.write(feature_str + "\n")
            f_out.close()
        elif index_type=="autoencoder":
            pass

    def export_key_values(self,field,save_path=""):
        data=self.get_data()
        list_item=[]
        for key in tqdm(data.keys()):
            content = data[key]
            if field in content.keys():
                v = content[field][()]
            else:
                v = ""
            if type(v)!=str:
                v=v.decode("utf-8",errors="ignore")
            list_item.append({"key":key,"value":v})
        if save_path!="":
            write_csv(save_path=save_path,list_rows=list_item)
        return list_item


    def search_obj_index(self,index_file,features):
        from mmkfeatures.image.color_descriptor import ColorDescriptor
        from mmkfeatures.image.image_searcher import Searcher
        # initialize the image descriptor
        print("Searching index....")
        # perform the search
        searcher = Searcher(index_file)
        results = searcher.search(features)
        # display the query
        # cv2.imshow("Query", query)
        # cv2.waitKey(0)
        return results

    def get_content_by_id(self,cid):
        return self.get_data()[cid]




import os
# import traceback
from typing import Dict, Union

import cv2
import mmkfeatures.video.models.pwc.pwc_src.utils.flow_viz as flow_viz
import numpy as np
import torch
import torchvision.transforms as transforms
from mmkfeatures.video.models.pwc.pwc_src.pwc_net import PWCNet
from mmkfeatures.video.models.pwc.transforms.transforms import (PILToTensor, ResizeImproved,
                                              ToFloat, ToTensorWithoutScaling)
from tqdm import tqdm
from mmkfeatures.video.utils.utils import (action_on_extraction, form_list_from_user_input,
                         reencode_video_with_diff_fps)

current_path = os.path.dirname(os.path.realpath(__file__)).replace("\\pwc","")

PWC_MODEL_PATH = current_path+ '/pwc/checkpoints/pwc_net_sintel.pt'

class ExtractPWC(torch.nn.Module):

    def __init__(self, args):
        super(ExtractPWC, self).__init__()
        self.feature_type = args.feature_type
        self.path_list = form_list_from_user_input(args)
        self.model_path = PWC_MODEL_PATH
        self.batch_size = args.batch_size
        self.extraction_fps = args.extraction_fps
        self.resize_to_smaller_edge = args.resize_to_smaller_edge
        self.side_size = args.side_size
        if self.side_size is not None:
            self.transforms = transforms.Compose([
                transforms.ToPILImage(),
                ResizeImproved(self.side_size, self.resize_to_smaller_edge),
                PILToTensor(),
                ToFloat(),
            ])
        else:
            self.transforms = transforms.Compose([
                ToTensorWithoutScaling()
            ])
        self.show_pred = args.show_pred
        self.keep_tmp_files = args.keep_tmp_files
        self.extraction_fps = args.extraction_fps
        self.on_extraction = args.on_extraction
        self.tmp_path = os.path.join(args.tmp_path, self.feature_type)
        self.output_path = os.path.join(args.output_path, self.feature_type)
        self.progress = tqdm(total=len(self.path_list))

    def forward(self, indices: torch.LongTensor):
        '''
        Arguments:
            indices {torch.LongTensor} -- indices to self.path_list
        '''
        device = indices.device
        model = self.load_model(device)

        for idx in indices:
            # when error occurs might fail silently when run from torch data parallel
            try:
                feats_dict = self.extract(device, model, self.path_list[idx])
                action_on_extraction(feats_dict, self.path_list[idx], self.output_path, self.on_extraction)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except Exception as e:
                # prints only the last line of an error. Use `traceback.print_exc()` for the whole traceback
                # traceback.print_exc()
                print(e)
                print(f'Extraction failed at: {self.path_list[idx]} with error (↑). Continuing extraction')

            # update tqdm progress bar
            self.progress.update()

    def extract(self, device: torch.device, model: torch.nn.Module,
                video_path: Union[str, None] = None) -> Dict[str, np.ndarray]:
        '''The extraction call. Made to clean the forward call a bit.

        Arguments:
            device {torch.device}
            model {torch.nn.Module}

        Keyword Arguments:
            video_path {Union[str, None]} -- if you would like to use import it and use it as
                                             "path -> model"-fashion (default: {None})

        Returns:
            Dict[str, np.ndarray]: 'features_name', 'fps', 'timestamps_ms'
        '''
        def _run_on_a_batch(flow_frames, batch, model, device):
            batch = torch.cat(batch).to(device)

            with torch.no_grad():
                flow = model(batch[:-1], batch[1:])
                # upadding only before saving because np.concat will not work if the img is unpadded
                flow_frames.extend(flow.tolist())
                # show optical flow along with rgb frames
                if self.show_pred:
                    self.show_flow_for_every_pair_of_frames(flow, batch)

        # take the video, change fps and save to the tmp folder
        if self.extraction_fps is not None:
            video_path = reencode_video_with_diff_fps(video_path, self.tmp_path, self.extraction_fps)

        cap = cv2.VideoCapture(video_path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        timestamps_ms = []
        batch = []
        flow_frames = []

        # sometimes when the target fps is 1 or 2, the first frame of the reencoded video is missing
        # and cap.read returns None but the rest of the frames are ok. timestep is 0.0 for the 2nd frame in
        # this case
        first_frame = True
        while cap.isOpened():
            frame_exists, rgb = cap.read()

            if first_frame:
                first_frame = False
                if frame_exists is False:
                    continue

            if frame_exists:
                timestamps_ms.append(cap.get(cv2.CAP_PROP_POS_MSEC))
                # preprocess the image
                rgb = cv2.cvtColor(rgb, cv2.COLOR_BGR2RGB)
                rgb = self.transforms(rgb)
                rgb = rgb.unsqueeze(0)

                batch.append(rgb)

                # - 1 is used because we need B+1 frames to calculate B frames
                if len(batch) - 1 == self.batch_size:
                    _run_on_a_batch(flow_frames, batch, model, device)
                    # leaving the last element to calculate flow between it and the first element
                    batch = [batch[-1]]
            else:
                if len(batch) > 1:
                    _run_on_a_batch(flow_frames, batch, model, device)
                cap.release()
                break

        # removes the video with different fps to preserve disk space
        if (self.extraction_fps is not None) and (not self.keep_tmp_files):
            os.remove(video_path)

        features_with_meta = {
            self.feature_type: np.array(flow_frames),
            'fps': np.array(fps),
            'timestamps_ms': np.array(timestamps_ms)
        }

        return features_with_meta

    def show_flow_for_every_pair_of_frames(self, out: torch.Tensor, batch: torch.Tensor):
        '''Shows the resulting flow as well as the first frame

        Args:
            out (torch.Tensor): the output of the model
            batch (torch.Tensor): the stack of rgb inputs
        '''
        for idx, flow in enumerate(out):
            img = batch[idx].permute(1, 2, 0).cpu().numpy()
            flow = flow.permute(1, 2, 0).cpu().numpy()
            flow = flow_viz.flow_to_image(flow)
            img_flow = np.concatenate([img, flow], axis=0)
            cv2.imshow('Press any key to see the next frame...', img_flow[:, :, [2, 1, 0]] / 255.0)
            cv2.waitKey()

    def load_model(self, device: torch.device) -> torch.nn.Module:
        '''Defines the models, loads checkpoints, sends them to the device.

        Args:
            device (torch.device): The device

        Returns:
            torch.nn.Module: flow extraction module.
        '''
        model = PWCNet()
        model.load_state_dict(torch.load(self.model_path, map_location='cpu'))
        model = model.to(device)
        model.eval()
        return model



import os
import pickle

import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import tensorflow as tf
from sklearn.model_selection import train_test_split
from keras.layers import *
from keras.models import *
from keras.losses import *
from keras.optimizers import *
from keras.callbacks import ModelCheckpoint
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

class CnnNode:

    def __init__(self,root_path,name):
        # load datasets
        self.name = name
        self.root_path=root_path
        self.model_root_path=self.root_path+"/"+name
        self.data_path=self.model_root_path+"/data"
        self.model_path=self.model_root_path+f"/{name}.h5"
        if not os.path.exists(self.model_root_path):
            os.mkdir(self.model_root_path)


    # Converts numbers to 6 figure strings
    # e.g. '24' to '000024', '357' to '000357'
    def to_6sf(self,n):
        new = str(n)
        new = '0' * (6 - len(new)) + new
        return new

    def build(self,n_epochs=10,train_dir="",ext=".jpeg",use_file_names=False):
        if train_dir=="":
            train_dir=self.data_path
        _, labels, _ = next(os.walk(train_dir))
        pickle.dump(labels,open(self.model_root_path+"/labels.pickle","wb"))
        print('Labels :', labels)
        x_all = []
        y_all = []
        for x in labels:

            if use_file_names:
                for file in os.listdir(train_dir + "/" + x ):
                    X = Image.open(train_dir + "/" + x + '/' + file)
                    X = np.array(X)
                    X=np.resize(X,(64,64,3))
                    X = np.reshape(X, (X.shape[0], X.shape[1], 3))
                    x_all.append(X)
                    y_all.append(x)
            else:
                num_images = len(os.listdir(train_dir + "/" + x + '/'))
                for i in range(num_images):
                    X = Image.open(train_dir + "/" + x + '/' + self.to_6sf(i) + ext)
                    X = np.array(X)
                    X = np.reshape(X, (X.shape[0], X.shape[1], 1))
                    x_all.append(X)
                    y_all.append(x)
        x_all = np.array(x_all)
        y_all = np.array(y_all)

        print('Total Samples :', len(x_all))

        ## preprocess datasets

        # Prepare Encoder-Decoder Dict
        encode_y = dict()
        decode_y = dict()
        for x in enumerate(labels):
            encode_y[x[1]] = x[0]
            decode_y[x[0]] = x[1]
        print('Encoder :', encode_y)
        print('Decoder :', decode_y)

        # Apply Encoder dict
        for x in range(len(x_all)):
            y_all[x] = encode_y[y_all[x]]
        y_all = np.array(y_all, dtype=np.int16)

        ## Split to Train & Test
        x_train, x_test, y_train, y_test = train_test_split(x_all, y_all, test_size=0.2, shuffle=True)
        print('Total Train Samples :', len(x_train))
        print('Total Test Samples :', len(x_test))

        # Model
        input_shape = x_all[0].shape
        print('Image Shape :', input_shape)

        model = Sequential()
        model.add(Input(shape=input_shape))
        model.add(Conv2D(64, kernel_size=(3, 3), padding='same', activation='relu'))
        model.add(Conv2D(40, kernel_size=(3, 3), padding='same', activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))
        model.add(Conv2D(32, kernel_size=(3, 3), padding='same', activation='relu'))
        model.add(Conv2D(16, kernel_size=(3, 3), padding='same', activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))
        model.add(Flatten())
        model.add(Dense(164, activation='relu'))
        model.add(Dense(128, activation='relu'))
        model.add(Dense(64, activation='relu'))
        model.add(Dense(42, activation='relu'))
        model.add(Dense(len(labels), activation='softmax'))

        model.compile(optimizer='adam',
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])

        model.summary()

        ## train model
        checkpoint = ModelCheckpoint(self.model_path)

        history = model.fit(x_all, y_all, epochs=n_epochs,
                  validation_data=(x_test, y_test), callbacks=[checkpoint])
        print(history.history)
        pickle.dump(history.history,open(self.model_root_path+"/history.pickle","wb"))

        # labels = pickle.load(open(self.model_root_path + "/labels.pickle", "rb"))

        # Load pretrained model
        model = load_model(self.model_path)
        loss, acc = model.evaluate(x_test, y_test)
        print('Test Loss :', loss)
        print('Test Accuracy :', acc)

    def show_history(self):
        if os.path.exists(self.model_root_path+"/history.pickle"):
            history=pickle.load(open(self.model_root_path+"/history.pickle","rb"))

            print(history)
            metrics=list(history.keys())
            c=len(history["accuracy"])
            f_out=open(self.model_root_path+"/history.csv","w",encoding='utf-8')
            print("\t".join(metrics))
            f_out.write("\t".join(metrics)+"\n")
            for i in range(c):
                line=""
                for l in metrics:
                    line+=str(history[l][i])+"\t"
                print(line)
                f_out.write(line+"\n")
            f_out.close()


    def predict(self,img,show_fig=False,channels=1):
        labels = pickle.load(open(self.model_root_path + "/labels.pickle", "rb"))
        # Prepare Encoder-Decoder Dict
        encode_y = dict()
        decode_y = dict()
        for x in enumerate(labels):
            encode_y[x[1]] = x[0]
            decode_y[x[0]] = x[1]
        print('Encoder :', encode_y)
        print('Decoder :', decode_y)

        # Load pretrained model
        model = load_model(self.model_path)

        # Predict
        # IMAGE_INDEX = 100
        # print("x_test.shape", x_test.shape)
        # img = x_test[IMAGE_INDEX]
        if show_fig:
            plt.title("origin")
            plt.imshow(img.squeeze())
            plt.show()

        # print('Actual :', decode_y[y_test[IMAGE_INDEX]])

        img = np.reshape(img, (1, 64, 64, channels))
        predicted_obj = model.predict(img)

        predicted_obj1 = np.argmax(predicted_obj)
        predicted_label = decode_y[predicted_obj1]
        print('Predicted :', predicted_label)
        return predicted_label




import decimal
import numpy
import math
import logging

def round_half_up(number):
    return int(decimal.Decimal(number).quantize(decimal.Decimal('1'), rounding=decimal.ROUND_HALF_UP))


def rolling_window(a, window, step=1):
    # http://ellisvalentiner.com/post/2017-03-21-np-strides-trick
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return numpy.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)[::step]

def framesig(sig, frame_len, frame_step, winfunc=lambda x: numpy.ones((x,)), stride_trick=True):
    """Frame a signal into overlapping frames.

    :param sig: the audio signal to frame.
    :param frame_len: length of each frame measured in samples.
    :param frame_step: number of samples after the start of the previous frame that the next frame should begin.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied.
    :param stride_trick: use stride trick to compute the rolling window and window multiplication faster
    :returns: an array of frames. Size is NUMFRAMES by frame_len.
    """
    slen = len(sig)
    frame_len = int(round_half_up(frame_len))
    frame_step = int(round_half_up(frame_step))
    if slen <= frame_len:
        numframes = 1
    else:
        numframes = 1 + int(math.ceil((1.0 * slen - frame_len) / frame_step))

    padlen = int((numframes - 1) * frame_step + frame_len)

    zeros = numpy.zeros((padlen - slen,))
    padsignal = numpy.concatenate((sig, zeros))
    if stride_trick:
        win = winfunc(frame_len)
        frames = rolling_window(padsignal, window=frame_len, step=frame_step)
    else:
        indices = numpy.tile(numpy.arange(0, frame_len), (numframes, 1)) + numpy.tile(
            numpy.arange(0, numframes * frame_step, frame_step), (frame_len, 1)).T
        indices = numpy.array(indices, dtype=numpy.int32)
        frames = padsignal[indices]
        win = numpy.tile(winfunc(frame_len), (numframes, 1))

    return frames * win


def deframesig(frames, siglen, frame_len, frame_step, winfunc=lambda x: numpy.ones((x,))):
    """Does overlap-add procedure to undo the action of framesig.

    :param frames: the array of frames.
    :param siglen: the length of the desired signal, use 0 if unknown. Output will be truncated to siglen samples.
    :param frame_len: length of each frame measured in samples.
    :param frame_step: number of samples after the start of the previous frame that the next frame should begin.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied.
    :returns: a 1-D signal.
    """
    frame_len = round_half_up(frame_len)
    frame_step = round_half_up(frame_step)
    numframes = numpy.shape(frames)[0]
    assert numpy.shape(frames)[1] == frame_len, '"frames" matrix is wrong size, 2nd dim is not equal to frame_len'

    indices = numpy.tile(numpy.arange(0, frame_len), (numframes, 1)) + numpy.tile(
        numpy.arange(0, numframes * frame_step, frame_step), (frame_len, 1)).T
    indices = numpy.array(indices, dtype=numpy.int32)
    padlen = (numframes - 1) * frame_step + frame_len

    if siglen <= 0: siglen = padlen

    rec_signal = numpy.zeros((padlen,))
    window_correction = numpy.zeros((padlen,))
    win = winfunc(frame_len)

    for i in range(0, numframes):
        window_correction[indices[i, :]] = window_correction[
                                               indices[i, :]] + win + 1e-15  # add a little bit so it is never zero
        rec_signal[indices[i, :]] = rec_signal[indices[i, :]] + frames[i, :]

    rec_signal = rec_signal / window_correction
    return rec_signal[0:siglen]

def magspec(frames, NFFT):
    """Compute the magnitude spectrum of each frame in frames. If frames is an NxD matrix, output will be Nx(NFFT/2+1).

    :param frames: the array of frames. Each row is a frame.
    :param NFFT: the FFT length to use. If NFFT > frame_len, the frames are zero-padded.
    :returns: If frames is an NxD matrix, output will be Nx(NFFT/2+1). Each row will be the magnitude spectrum of the corresponding frame.
    """
    if numpy.shape(frames)[1] > NFFT:
        logging.warn(
            'frame length (%d) is greater than FFT size (%d), frame will be truncated. Increase NFFT to avoid.',
            numpy.shape(frames)[1], NFFT)
    complex_spec = numpy.fft.rfft(frames, NFFT)
    return numpy.absolute(complex_spec)


def powspec(frames, NFFT):
    """Compute the power spectrum of each frame in frames. If frames is an NxD matrix, output will be Nx(NFFT/2+1).

    :param frames: the array of frames. Each row is a frame.
    :param NFFT: the FFT length to use. If NFFT > frame_len, the frames are zero-padded.
    :returns: If frames is an NxD matrix, output will be Nx(NFFT/2+1). Each row will be the power spectrum of the corresponding frame.
    """
    return 1.0 / NFFT * numpy.square(magspec(frames, NFFT))


def logpowspec(frames, NFFT, norm=1):
    """Compute the log power spectrum of each frame in frames. If frames is an NxD matrix, output will be Nx(NFFT/2+1).

    :param frames: the array of frames. Each row is a frame.
    :param NFFT: the FFT length to use. If NFFT > frame_len, the frames are zero-padded.
    :param norm: If norm=1, the log power spectrum is normalised so that the max value (across all frames) is 0.
    :returns: If frames is an NxD matrix, output will be Nx(NFFT/2+1). Each row will be the log power spectrum of the corresponding frame.
    """
    ps = powspec(frames, NFFT);
    ps[ps <= 1e-30] = 1e-30
    lps = 10 * numpy.log10(ps)
    if norm:
        return lps - numpy.max(lps)
    else:
        return lps


def preemphasis(signal, coeff=0.95):
    """perform preemphasis on the input signal.

    :param signal: The signal to filter.
    :param coeff: The preemphasis coefficient. 0 is no filter, default is 0.95.
    :returns: the filtered signal.
    """
    return numpy.append(signal[0], signal[1:] - coeff * signal[:-1])



# import the necessary packages
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv2DTranspose
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Reshape
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model
from tensorflow.keras import backend as K
import numpy as np

class ConvAutoencoder:
	@staticmethod
	def build(width, height, depth, filters=(32, 64), latentDim=16):
		# initialize the input shape to be "channels last" along with
		# the channels dimension itself
		# channels dimension itself
		inputShape = (height, width, depth)
		chanDim = -1
		# define the input to the encoder
		inputs = Input(shape=inputShape)
		x = inputs
		# loop over the number of filters
		for f in filters:
			# apply a CONV => RELU => BN operation
			x = Conv2D(f, (3, 3), strides=2, padding="same")(x)
			x = LeakyReLU(alpha=0.2)(x)
			x = BatchNormalization(axis=chanDim)(x)
		# flatten the network and then construct our latent vector
		volumeSize = K.int_shape(x)
		x = Flatten()(x)
		latent = Dense(latentDim, name="encoded")(x)

		# start building the decoder model which will accept the
		# output of the encoder as its inputs
		x = Dense(np.prod(volumeSize[1:]))(latent)
		x = Reshape((volumeSize[1], volumeSize[2], volumeSize[3]))(x)
		# loop over our number of filters again, but this time in
		# reverse order
		for f in filters[::-1]:
			# apply a CONV_TRANSPOSE => RELU => BN operation
			x = Conv2DTranspose(f, (3, 3), strides=2,
				padding="same")(x)
			x = LeakyReLU(alpha=0.2)(x)
			x = BatchNormalization(axis=chanDim)(x)
		# apply a single CONV_TRANSPOSE layer used to recover the
		# original depth of the image
		x = Conv2DTranspose(depth, (3, 3), padding="same")(x)
		outputs = Activation("sigmoid", name="decoded")(x)
		# construct our autoencoder model
		autoencoder = Model(inputs, outputs, name="autoencoder")
		# return the autoencoder model
		return autoencoder


import os

import numpy as np
from keras.applications.inception_resnet_v2 import InceptionResNetV2
from keras.applications.inception_resnet_v2 import preprocess_input as resnet_preprocessor
from keras.applications.inception_v3 import InceptionV3
from keras.applications.inception_v3 import preprocess_input as inception_preprocessor
from keras.applications.vgg19 import VGG19
from keras.applications.vgg19 import preprocess_input as vgg19_preprocessor
from keras.preprocessing import image as image_preprocessor

from mmkfeatures.image.extractor import Extractor


class DeepExtractor(Extractor):
    
    def __init__(self, base_route, model_name: str, size=224, batch_size=128):
        
        super().__init__(base_route=base_route, size=size, batch_size=batch_size)
        
        self.model = None
        self.file_writer = None
        self.model_preprocess = None
        self.model_name = model_name
        
        self._set_extractor_model(self.model_name)
    
    def _set_extractor_model(self, model_name: str) -> None:
        if model_name == "inception_v3":
            self.model = InceptionV3(include_top=False, weights="imagenet", input_shape=self.image_shape)
            self.model_preprocess = inception_preprocessor
        elif model_name == "inception_resnet_v2":
            self.model = InceptionResNetV2(include_top=False, weights="imagenet", input_shape=self.image_shape)
            self.model_preprocess = resnet_preprocessor
        elif model_name == "vgg19":
            self.model = VGG19(include_top=False, weights="imagenet", input_shape=self.image_shape)
            self.model_preprocess = vgg19_preprocessor
        else:
            raise Exception("Invalid pre-trained Keras Application")
    
    def extract(self, image_route: str) -> np.ndarray:
        image = image_preprocessor.load_img(image_route, target_size=(self.width, self.height))
        image = np.expand_dims(image_preprocessor.img_to_array(image), axis=0)
        preprocessed_img = self.model_preprocess(image)
        
        return self.model.predict(preprocessed_img).flatten()
    
    def _find_features_size(self) -> int:
        example_image_route = os.path.join(self.base_route, self.directory_iterator.filenames[0])
        return len(self.extract(image_route=example_image_route))



#CMU Multimodal SDK, CMU Multimodal Model SDK

#Tensor Fusion Network for Multimodal Sentiment Analysis, Amir Zadeh, Minghai Chen, Soujanya Poria, Erik Cambria, Louis-Philippe Morency - https://arxiv.org/pdf/1707.07250.pdf

#in_modalities: is a list of inputs from each modality - the first dimension of all the modality inputs must be the same, it will be the batch size. The second dimension is the feature dimension. There are a total of n modalities.

#out_dimension: the output of the tensor fusion

import torch
import time
from torch import nn
import torch.nn.functional as F
from six.moves import reduce

class TensorFusion(nn.Module):

	def __init__(self,in_dimensions,out_dimension):
		super(TensorFusion, self).__init__()
		self.tensor_size=reduce(lambda x, y: x*y, in_dimensions)
		self.linear_layer=nn.Linear(self.tensor_size,out_dimension)
		self.in_dimensions=in_dimensions
		self.out_dimension=out_dimension

	def __call__(self,in_modalities):
		return self.fusion(in_modalities)

	def fusion(self,in_modalities):

		bs=in_modalities[0].shape[0]
		tensor_product=in_modalities[0]
		
		#calculating the tensor product
		
		for in_modality in in_modalities[1:]:
			tensor_product=torch.bmm(tensor_product.unsqueeze(2),in_modality.unsqueeze(1))
			tensor_product=tensor_product.view(bs,-1)
		
		return self.linear_layer(tensor_product)

	def forward(self, x):
		print("Not yet implemented for nn.Sequential")
		exit(-1)

if __name__=="__main__":

	print("This is a module and hence cannot be called directly ...")
	print("A toy sample will now run ...")
	
	from torch.autograd import Variable
	import torch.nn.functional as F
	import numpy
	# in=[batch size, feature dimension]
	inputx=Variable(torch.Tensor(numpy.zeros([32,40])),requires_grad=True)

	print(inputx)

	inputy=Variable(torch.Tensor(numpy.array(numpy.zeros([32,12]))),requires_grad=True)
	inputz=Variable(torch.Tensor(numpy.array(numpy.zeros([32,20]))),requires_grad=True)
	modalities=[inputx,inputy,inputz]

	print("InputX:")
	print(inputx)
	
	fmodel=TensorFusion([40,12,20],100)

	# output =  [batch size, feature dimension]
	out=fmodel(modalities)

	print("Output")
	print(out[0])
	print(out.shape)
	print("Toy sample finished ...")










import os
import multiprocessing
import ffmpeg

if __name__ == '__main__':
    multiprocessing.freeze_support()
    '''
    
    import ffmpeg
    stream = ffmpeg.input(root_path+'/videos/a.mp4')
    stream = ffmpeg.hflip(stream)
    stream = ffmpeg.output(stream, root_path+'/videos/a_out.mp4')
    ffmpeg.run(stream)
    '''

    for line in open("../dataset/input.csv",'r',encoding='utf-8').readlines():
        ls=line.strip().split(",")
        if os.path.exists(ls[1]):
            os.remove(ls[1])

    os.system("python extract.py --csv=../dataset/input.csv --type=2d --batch_size=4 --num_decoding_thread=4")




import math
import os

from tensorflow.python.keras.applications.vgg19 import VGG19
from tensorflow.python.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.python.keras.layers import Dense
from tensorflow.python.keras.layers import GlobalAveragePooling2D
from tensorflow.python.keras.models import Model
from tensorflow.keras.optimizers import SGD
from tensorflow.python.keras.preprocessing import image


class ImageModel(object):
    def __init__(self, base_route, train_folder="train", validation_folder="val", epochs=10, fine_tune: bool = False):
        self.__model = None
        self.__base_model = None
        self.__width = self.__height = 64
        self.__train_route = os.path.join(base_route, train_folder)
        self.__validation_route = os.path.join(base_route, validation_folder)
        
        self.__fine_tuning = fine_tune
        
        self.__epochs = epochs
        self.__batch_size = 256
        
        self.__model_route = "model.h5"
        self.__early_stop = EarlyStopping(monitor='val_acc', min_delta=0, patience=10, verbose=1, mode='auto')
        self.__checkpoint = self._get_model_checkpoint()
        
        self.train_size = 0
        self.validation_size = 0
        self.train_steps = 0
        self.validation_steps = 0
    
    def train(self):
        train_directory_iterator = self._get_directory_iterator(self.__train_route)
        validation_directory_iterator = self._get_directory_iterator(self.__validation_route)
        
        self.train_size = train_directory_iterator.samples
        self.validation_size = validation_directory_iterator.samples
        
        self._build_model(train_directory_iterator.num_classes)
        
        if self.__fine_tuning:
            self._set_fine_tune()
        else:
            self._set_transfer_learning()
        
        self.__model.fit_generator(
            train_directory_iterator,
            steps_per_epoch=self.train_steps,
            epochs=self.__epochs,
            validation_data=validation_directory_iterator,
            validation_steps=self.validation_steps,
            callbacks=[self.__checkpoint, self.__early_stop]
        )
        
        self.__model.save(self.__model_route)
        
        metrics = self.__model.evaluate_generator(train_directory_iterator)
        
        return metrics
    
    def _build_model(self, num_classes: int):
        self.__base_model = VGG19(weights='imagenet', include_top=False, input_shape=(self.__width, self.__height, 3))
        
        x = self.__base_model.output
        x = GlobalAveragePooling2D()(x)
        x = Dense(512, activation='relu')(x)
        predictions = Dense(num_classes, activation='softmax')(x)
        
        self.__model = Model(inputs=self.__base_model.input, outputs=predictions)
    
    def _set_transfer_learning(self):
        for layer in self.__base_model.layers:
            layer.trainable = False
        
        self.__model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    
    def _set_fine_tune(self):
        layers_to_freeze = int(len(self.__model.layers) * 0.9)
        
        for layer in self.__model.layers[:layers_to_freeze]:
            layer.trainable = False
        for layer in self.__model.layers[layers_to_freeze:]:
            layer.trainable = True
        
        self.__model.compile(
            optimizer=SGD(lr=0.0001, momentum=0.9),
            loss='categorical_crossentropy',
            metrics=['accuracy']
        )
    
    def _get_model_checkpoint(self):
        return ModelCheckpoint(
            self.__model_route,
            monitor='val_acc',
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
            mode='auto',
            period=1
        )
    
    def _get_directory_iterator(self, route):
        image_generator = image.ImageDataGenerator(rescale=1.0 / 255)
        
        return image_generator.flow_from_directory(
            directory=route,
            target_size=(self.__width, self.__height),
            batch_size=self.batch_size,
            class_mode="categorical"
        )
    
    @property
    def train_size(self):
        return self.__train_size
    
    @train_size.setter
    def train_size(self, train_size):
        self.__train_size = train_size
        self.train_steps = math.ceil(self.train_size / self.batch_size)
    
    @property
    def validation_size(self):
        return self.__validation_size
    
    @validation_size.setter
    def validation_size(self, validation_size):
        self.__validation_size = validation_size
        self.validation_steps = math.ceil(self.validation_size / self.batch_size)
    
    @property
    def batch_size(self):
        return self.__batch_size
    
    @batch_size.setter
    def batch_size(self, batch_size):
        self.__batch_size = batch_size
        self.train_steps = math.ceil(self.train_size / self.batch_size)
        self.validation_steps = math.ceil(self.validation_size / self.batch_size)



import math
import numpy as np

def ZeroCR(waveData,frameSize,overLap):
    wlen = len(waveData)
    step = frameSize - overLap
    frameNum = int(math.ceil(wlen*1.0/step))
    zcr = np.zeros((frameNum,1))
    for i in range(frameNum):
        curFrame = waveData[np.arange(i*step, min(i*step+frameSize,wlen))]
        #To avoid DC bias, usually we need to perform mean subtraction on each frame
        curFrame = curFrame - np.mean(curFrame) # zero-justified
        zcr[i] = sum(curFrame[0:-1]*curFrame[1:]<=0)
    return zcr



import torch
from torch import nn
from torch.autograd import Variable
import torch.nn.functional as F
from LSTHM import LSTHM
import numpy


#in=40, 3 modalities,4 attentions
full_in=numpy.array(numpy.zeros([32,40]))
inputx=Variable(torch.Tensor(full_in),requires_grad=True)
full_in=numpy.array(numpy.zeros([32,12]))
inputy=Variable(torch.Tensor(full_in),requires_grad=True)
full_in=numpy.array(numpy.zeros([32,12]))
inputc=Variable(torch.Tensor(full_in),requires_grad=True)
full_in=numpy.array(numpy.zeros([32,20]))
inputz=Variable(torch.Tensor(full_in),requires_grad=True)

fmodel=LSTHM(12,40,20)
c,h=fmodel.step(inputx,inputc,inputy,inputz)
print(c.shape,h.shape)
print(c)

#a=numpy.array([[1,2,3],[4,5,6]])
#b=Variable(torch.Tensor(a))
#c=b.repeat(1,2)
#print(c.shape)
#print(F.softmax(c,dim=1))
#print(c.view(2,2,3)[0,1,:])





import torch as th
import math
import numpy as np
from mmkfeatures.video.model import get_model_p
from mmkfeatures.video.preprocessing import Preprocessing
import torch.nn.functional as F
import ffmpeg

class VideoFeatureWrapper:
    def __init__(self):
        self.framerate = 1
        self.size = 112
        self.centercrop = False
        self.model_type = "2d"  # 3d
        self.batch_size = 4
        self.l2_normalize = 1
        self.half_precision = 1
        self.resnext101_model_path = "model/resnext101.pth"
        self.num_decoding_thread = 4

    def _get_video_dim(self,video_path):
        probe = ffmpeg.probe(video_path)
        video_stream = next((stream for stream in probe['streams']
                             if stream['codec_type'] == 'video'), None)
        width = int(video_stream['width'])
        height = int(video_stream['height'])
        return height, width

    def _get_output_dim(self,h, w):
        if isinstance(self.size, tuple) and len(self.size) == 2:
            return self.size
        elif h >= w:
            return int(h * self.size / w), self.size
        else:
            return self.size, int(w * self.size / h)

    def load_video(self,video_path):
        try:
            # video_path = ""
            output_file = ""
            h, w = self._get_video_dim(video_path)
        except:
            print('ffprobe failed at: {}'.format(video_path))

        height, width = self._get_output_dim(h, w)
        cmd = (
            ffmpeg
                .input(video_path)
                .filter('fps', fps=self.framerate)
                .filter('scale', width, height)
        )
        if self.centercrop:
            x = int((width - self.size) / 2.0)
            y = int((height - self.size) / 2.0)
            cmd = cmd.crop(x, y, self.size, self.size)
        out, _ = (
            cmd.output('pipe:', format='rawvideo', pix_fmt='rgb24')
                .run(capture_stdout=True, quiet=True)
        )
        if self.centercrop and isinstance(self.size, int):
            height, width = self.size, self.size
        video = np.frombuffer(out, np.uint8).reshape([-1, height, width, 3])
        video = th.from_numpy(video.astype('float32'))
        video = video.permute(0, 3, 1, 2)
        return video

    def extract_video_features(self,video_file,output_file=None):
        #video_file = "../dataset/videos/a.mp4"
        #output_file = "../dataset/output/single_features.npy"

        preprocess = Preprocessing(self.model_type)
        model = get_model_p(self.model_type, self.resnext101_model_path)

        with th.no_grad():
            data = self.load_video(video_file)

            video = data.squeeze()
            print("len(video.shape)=", len(video.shape))

            video = preprocess(video)

            n_chunk = len(video)
            print("n_chunk = ", n_chunk)
            print("batch size = ", self.batch_size)
            features = th.cuda.FloatTensor(n_chunk, 2048).fill_(0)
            n_iter = int(math.ceil(n_chunk / float(self.batch_size)))
            print("number of inter. = ", n_iter)
            for i in range(n_iter):
                min_ind = i * self.batch_size
                max_ind = (i + 1) * self.batch_size
                video_batch = video[min_ind:max_ind].cuda()
                batch_features = model(video_batch)
                if self.l2_normalize:
                    batch_features = F.normalize(batch_features, dim=1)
                features[min_ind:max_ind] = batch_features
            features = features.cpu().numpy()
            print("features.shape = ", features.shape)
            if self.half_precision:
                features = features.astype('float16')
            print("save...", output_file)
            # print(features.shape)
            if output_file!=None:
                np.save(output_file, features)
            print()
        return features


import numpy as np
import math
import string
import pickle
from functools import reduce
from mmkfeatures.text.structures import Document, Sentence, Token

class AbstractFeaturizer():
    """
    Abstract class for turning text into features.
    Attributes
    ----------
    ignore_tokens: list of str
        List of tokens to ignore in Featurization computing. E.g.: <SOS> and <EOS> tokens (default).
    ignore_punctuation: boolean
        Whether to ignore punctiation as features. Defaults to True. Uses string.punctuation.
    lower_case: boolean
        Whether to turn all text to lowercase or keep case. Default to false (keeps case).
    word_indexes: dict of int
        A dictionary with the index of each word fit into the featurizer. This is used to compose the feature vector.
    """

    def __init__(self, ignore_tokens=["<SOS>","<EOS>"], ignore_punctuation=True, lower_case = False):
        self.ignore_tokens = ignore_tokens
        if ignore_punctuation:
            self.ignore_tokens+=[char for char in string.punctuation]
        self.lower_case=lower_case
        self.word_indexes={}
        self.index_to_word={}

    def fit(self, data):
        """
        Fits the data into the Featurizer.
        Abstract class only implements the exceptions.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document or list of strings.
        """

        if not (isinstance(data, Document) or isinstance(data, list)):
            raise(TypeError("Could not fit data because it is of wrong type. Type {} found. Please, insert it as a Document or list of strings".format(type(data))))

    def transform(self, data):
        """
        Transforms the data passed as input.
        Abstract class only implements the exceptions.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document, Sentence, list of strings or string.
        AttributeError
            Related to the vocabulary lenght. Happens if fit with empty data or not fit.
        """
        if not (isinstance(data, Document) or isinstance(data, Sentence) or isinstance(data, str) or isinstance(data, list)):
            raise(TypeError("Could not transform data because it is of wrong type. Type {} found. Please, insert it as a Document, list of strings, Sentence or string.".format(type(data))))
        if len(self.word_indexes) == 0:
            raise(AttributeError("Vocabulary length is zero. Maybe you forgot to fit the Featurizer?"))

    def fit_transform(self, data):
        """
        Fits and then transforms the data passed as input.
        """
        self.fit(data)
        return self.transform(data)

    def _transform_document(self):
        pass

    def _transform_sentence(self):
        pass

    def save_to_file(self, filename):
        """
        Saves the current status of the Featurizer to a pickled file. Useful for deploying.
        """
        if not filename.endswith(".p"):
            filename+=".p"
        with open(filename, 'wb') as f:
            pickle.dump(self.__dict__, f)
        print("{} saved successfully as a pickeld file. You can load it using 'load_from_file' function.")

    def load_from_file(self, filename):
        """
        Loads a pickled featurizer, updating the current featurizer to match the saved data.
        """
        with open(filename, 'rb') as f:
            tmp_dict = pickle.load(f)
        self.__dict__.update(tmp_dict)

class Bow(AbstractFeaturizer):
    """
    Bag of Words implementation for featurization. Based on AbstractFeaturizer class.
    """

    def __init__(self, ignore_tokens=["<SOS>","<EOS>"], ignore_punctuation=True, lower_case = False):
        super().__init__(ignore_tokens, ignore_punctuation, lower_case)

    def fit(self, data):
        """
        Fits the data into the Featurizer.
        Arguments
        ---------
        data: Document or list of string.
            The data to fit the featurizer.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document or list of strings.
        """
        super(Bow, self).fit(data)
        document_words = list(set(global_term_frequency(data, self.ignore_tokens, self.lower_case).keys()))
        for word_position in range(len(document_words)):
            word = document_words[word_position]
            self.word_indexes[word] = word_position

    def transform(self, data):
        """
        Transforms the data passed as input into a Bag of Words vector/matrix, depending on the input.
        Arguments
        ---------
        data: Document, Sentence, list of string or string.
            The data to fit the featurizer.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document, Sentence, list of strings or string.
        AttributeError
            Related to the vocabulary lenght. Happens if fit with empty data or not fit.
        """
        super(Bow, self).transform(data)
        if isinstance(data, Document) or isinstance(data, list):
            return self._transform_document(data)
        elif isinstance(data, Sentence) or isinstance(data, str):
            return self._transform_sentence(data)

    def _transform_document(self, data):
        if isinstance(data, Document):
            to_transform = data.sentences
        else:
            to_transform = data
        sentence_arrays = []
        for sentence in data:
            sentence_arrays.append(self._transform_sentence(sentence))
        return np.matrix(sentence_arrays)

    def _transform_sentence(self, data):
        if isinstance(data, Sentence):
            tokens = [token.get().lower() if self.lower_case else token.get() for token in data.tokens if token not in self.ignore_tokens]
        else:
            tokens = [token.lower() if self.lower_case else token for token in data.split()]
        word_array = np.zeros(len(self.word_indexes))
        for token in tokens:
            # Dismisses out of vocabulary tokens
            if token in self.word_indexes:
                token_index = self.word_indexes[token]
                word_array[token_index]+=1
        return word_array


class Tfidf(AbstractFeaturizer):
    """
    TF-IDF implementation for featurization. Based on AbstractFeaturizer class.
    Attributes:
    -----------
    idf_dict: dict of int.
        Dictionary containing the idf score for each word (available after fitting).
    num_documents: int.
        Number of documents used to fit the featurizer (important feature for score calculation).
    """

    def __init__(self, ignore_tokens=["<SOS>","<EOS>"], ignore_punctuation=True, lower_case=False):
        super().__init__(ignore_tokens, ignore_punctuation, lower_case)
        self.idf_dict = {}
        self.num_documents = 0

    def fit(self, data):
        """
        Fits the data into the Featurizer.
        Arguments
        ---------
        data: Document or list of string.
            The data to fit the featurizer.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document or list of strings.
        """
        super(Tfidf, self).fit(data)
        self.num_documents = len(data) if isinstance(data, list) else len(data.sentences)
        word_freq_per_document = self._compute_global_tf(data)
        self._compute_idf(word_freq_per_document)
        document_words = list(set(global_term_frequency(data, self.ignore_tokens, self.lower_case)))
        for word_position in range(len(document_words)):
            word = document_words[word_position]
            self.word_indexes[word] = word_position
            self.index_to_word[word_position] = word

    def transform(self, data):
        """
        Transforms the data passed as input into a tdf-idf vector/matrix, depending on the input.
        Arguments
        ---------
        data: Document, Sentence, list of string or string.
            The data to fit the featurizer.
        Exceptions
        ----------
        TypeError
            Related to data type. Expects Document, Sentence, list of strings or string.
        AttributeError
            Related to the vocabulary lenght. Happens if fit with empty data or not fit.
        """
        super(Tfidf, self).transform(data)
        if isinstance(data, Document) or isinstance(data, list):
            return self._transform_document(data)
        elif isinstance(data, Sentence) or isinstance(data, str):
            return self._transform_sentence(data)

    def _transform_document(self, data):
        if isinstance(data, Document):
            to_transform = data.sentences
        else:
            to_transform = data
        sentence_arrays = []
        for sentence in data:
            sentence_arrays.append(self._transform_sentence(sentence))
        return np.matrix(sentence_arrays)

    def _transform_sentence(self, data):
        if isinstance(data, Sentence):
            tokens = [token.get().lower() if self.lower_case else token.get() for token in data.tokens if token not in self.ignore_tokens]
        else:
            tokens = [token.lower() if self.lower_case else token for token in data.split()]
        word_array = np.zeros(len(self.word_indexes))
        sentence_tf_idf = self._compute_sentence_tf_idf(data)
        for token in tokens:
            # Dismisses out of vocabulary tokens
            if token in self.word_indexes:
                token_index = self.word_indexes[token]
                word_array[token_index] = sentence_tf_idf[token]
        return word_array

    def _compute_global_tf(self, data):
        """
        Computes global term frequency from data.
        """
        word_freq_per_document = {}
        if isinstance(data, Document):
            list_of_sentences = data.sentences
        else:
            list_of_sentences = data
        for sentence in list_of_sentences:
            words_in_sent = set()
            document_frequency = term_frequency(sentence, self.ignore_tokens, self.lower_case)
            for word in document_frequency:
                if not word in words_in_sent:
                    word_freq_per_document[word] = word_freq_per_document.get(word, 0)+1
                    words_in_sent.add(word)
        return word_freq_per_document

    def explain(self, tf_idf_array, summary=False):
        explained_results = {}
        if len(tf_idf_array.shape)>1:
            for sent_id, sent_scores in enumerate(tf_idf_array):
                explained_results[sent_id]={}
                for word_idx, idf in enumerate(sent_scores):
                    if summary:
                        if idf == 0.0:
                            continue
                    word = self.index_to_word[word_idx]
                    explained_results[sent_id][word]=idf
        else:
            explained_results[0]={}
            for word_idx, idf in enumerate(tf_idf_array):
                if summary:
                    if idf == 0.0:
                        continue
                word = self.index_to_word[word_idx]
                explained_results[0][word]=idf
        return explained_results


    def _compute_idf(self, word_freq_per_document):
        """
        Computes sentence(document) idf.
        """
        for word, frequency in word_freq_per_document.items():
            idf = math.log(float(1 + self.num_documents) / (1 + frequency))
            self.idf_dict[word]=idf

    def _compute_sentence_tf_idf(self, sentence):
        """
        Computes the tf_idf for a single sentence(document).
        """
        sentence_tf_idf = {}
        document_frequency = term_frequency(sentence, self.ignore_tokens, self.lower_case)
        total_words = sum(document_frequency.values())
        averaged_frequency = {k:(float(v)/total_words) for k,v in document_frequency.items()}
        for term, tf in averaged_frequency.items():
            # Out of vocabulary words are simply zeroed. They are going to be removed later either way.
            sentence_tf_idf[term] = tf*self.idf_dict.get(term, 0)
        return sentence_tf_idf


class Word2VecTransformer(AbstractFeaturizer):
    def fit(self):
        pass

    def transform(self):
        pass

    def fit_transform(self):
        pass

    def _transform_document(self):
        pass

    def _transform_sentence(self):
        pass

def global_term_frequency(document, ignore_tokens=["<SOS>","<EOS>"], lower_case = False):
    """
    Function to compute a list of terms and their frequency from a document (or set of documents). Used as auxiliary for other methods.
    Arguments:
    ----------
    document: list of string or Document class object.
        The "document" (or set of documents) to compute the global term frequency.
    ignore_tokens: list of str.
        Tokens to ignore in the term frequency computation.
    lower_case: boolean.
        Whether to be case sensitive or not. Defaults to False (case sensitive).
    """
    word_dict = {}
    if isinstance(document, Document):
        list_of_sentences = document.sentences
    else:
        list_of_sentences = document
    sentences_freqs = []
    for sentence in list_of_sentences:
        sentences_freqs.append(term_frequency(sentence, ignore_tokens, lower_case))
    word_dict = reduce_term_frequency(sentences_freqs)
    return word_dict

def term_frequency(sentence, ignore_tokens=["<SOS>","<EOS>"], lower_case = False):
    """
    Computes the term frequency for a single sentence (or document). Used as auxiliary for other methods.
    Arguments:
    ----------
    document: list of string or Document class object.
        The "document" (or set of documents) to compute the global term frequency.
    ignore_tokens: list of str.
        Tokens to ignore in the term frequency computation.
    lower_case: boolean.
        Whether to be case sensitive or not. Defaults to False (case sensitive).
    """
    word_dict = {}
    if isinstance(sentence, Sentence):
        words = [token.get().lower() if lower_case else token.get() for token in sentence.tokens if token not in ignore_tokens]
    else:
        words = [token.lower() if lower_case else token for token in sentence.split() if token not in ignore_tokens]
    for word in words:
        word_dict[word] = word_dict.get(word, 0)+1
    return word_dict

def reduce_term_frequency(list_of_word_dict):
    """
    Reduces a list of dicts to a single dictionary, summing all related values. Used to condense the results from the application of multimple term_frequency over several sentences. Returns a single 'str':int dict.
    Arguments:
    ----------
    list_of_word_dict: list of dict
        List of word dicts, containing pairs in 'str':int format, where the string is the word and the int is the frequency.
    """
    def reducer(accumulator, element):
        for key, value in element.items():
            accumulator[key] = accumulator.get(key, 0) + value
        return accumulator
    total_frequencies = reduce(reducer, list_of_word_dict, {})
    return total_frequencies
